#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017,2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import datetime
import urllib.error
import urllib.parse
import urllib.request

from gnuradio import gr
import pmt
import numpy


def parse_time(time):
    try:
        return datetime.datetime.fromisofromat(time)
    except AttributeError:
        # Workaround for Python version <3.7, which doesn't have
        # fromisoformat()
        return datetime.datetime.strptime(time, '%Y-%m-%dT%H:%M:%S')


class submit(gr.basic_block):
    """docstring for block submit"""
    def __init__(self, url, noradID, source,
                 longitude, latitude, initialTimestamp):
        gr.basic_block.__init__(
            self,
            name='submit',
            in_sig=[],
            out_sig=[])

        self.url = url
        self.request = {
            'noradID': noradID,
            'source': source,
            'locator': 'longLat',
            'longitude': str(
                abs(longitude)) + ('E' if longitude >= 0 else 'W'),
            'latitude': str(
                abs(latitude)) + ('N' if latitude >= 0 else 'S'),
            'version': '1.6.6',
            }
        self.initialTimestamp = (
            parse_time(initialTimestamp)
            if initialTimestamp != '' else None)
        self.startTimestamp = datetime.datetime.utcnow()

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)

    def handle_msg(self, msg_pmt):
        # Check that callsign and QTH have been entered
        if self.request['source'] == '':
            return
        if (self.request['longitude'] == 0.0
                and self.request['latitude'] == 0.0):
            return

        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return

        self.request['frame'] = bytes(pmt.u8vector_elements(msg)).hex().upper()

        now = datetime.datetime.utcnow()
        timestamp = (
            now - self.startTimestamp + self.initialTimestamp
            if self.initialTimestamp else now)
        self.request['timestamp'] = timestamp.isoformat()[:-3] + 'Z'

        params = urllib.parse.urlencode(self.request)
        try:
            f = urllib.request.urlopen(
                '{}?{}'.format(self.url, params),
                data=bytes(params, encoding='ascii'))
        except Exception as e:
            print('Error while submitting telemetry:', e)
            return
        reply = f.read()
        code = f.getcode()
        if code < 200 or code >= 300:
            print('Server error while submitting telemetry')
            print('Reply:')
            print(reply)
            print('URL:', f.geturl())
            print('HTTP code:', f.getcode())
            print('Info:')
            print(f.info())
        f.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import collections

from gnuradio import gr
import numpy
import pmt

from .kiss import *


class kiss_to_pdu(gr.sync_block):
    """docstring for block kiss_to_pdu"""
    def __init__(self, control_byte=True):
        gr.sync_block.__init__(
            self,
            name='kiss_to_pdu',
            in_sig=[numpy.uint8],
            out_sig=[])

        self.pdu = list()
        self.transpose = False
        self.control_byte = control_byte

        self.message_port_register_out(pmt.intern('out'))

    def work(self, input_items, output_items):
        for c in input_items[0]:
            if c == FEND:
                if (self.pdu
                        and (not self.control_byte or not self.pdu[0] & 0x0f)):
                    msg = bytes(self.pdu[1:]
                                if self.control_byte else self.pdu)
                    self.message_port_pub(
                        pmt.intern('out'),
                        pmt.cons(pmt.PMT_NIL,
                                 pmt.init_u8vector(len(msg), msg)))
                self.pdu = list()
            elif self.transpose:
                if c == TFEND:
                    self.pdu.append(FEND)
                elif c == TFESC:
                    self.pdu.append(FESC)
                self.transpose = False
            elif c == FESC:
                self.transpose = True
            else:
                self.pdu.append(c)

        return len(input_items[0])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr
import numpy as np
import pmt


def reflect_bytes(x):
    return np.fliplr(x[:x.size//8*8].reshape((-1, 8))).ravel()


def destuff(x):
    y = list()
    run = 0
    for i, bit in enumerate(x):
        if run == 5:
            if bit == 1:
                # Unexpected long run of ones
                return
            else:
                run = 0
        elif bit == 0:
            run = 0
            y.append(bit)
        elif bit == 1:
            run += 1
            y.append(bit)
    return np.array(y, dtype='uint8')


def descramble(x):
    y = np.concatenate((np.zeros(17, dtype='uint8'), x))
    z = y[:-17] ^ y[5:-12] ^ y[17:]
    return z


def nrzi_decode(x):
    return x ^ np.concatenate((np.zeros(1, dtype='uint8'), x[:-1])) ^ 1


class eseo_line_decoder(gr.basic_block):
    """docstring for block eseo_line_decoder"""
    def __init__(self):
        gr.basic_block.__init__(
            self,
            name='eseo_line_decoder',
            in_sig=[],
            out_sig=[])

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = np.array(pmt.u8vector_elements(msg), dtype='uint8')

        # Decode ESEO "line coding"
        packet = np.unpackbits(packet)
        packet = destuff(packet)
        if packet is None:
            return
        packet = descramble(packet)
        packet = nrzi_decode(packet)
        packet = reflect_bytes(packet)
        # Remove dummy padding
        cutbits = packet.size % 8
        if cutbits:
            packet = packet[:-cutbits]
        packet = np.packbits(packet)

        packet = bytes(packet)
        self.message_port_pub(pmt.intern('out'),
                              pmt.cons(pmt.car(msg_pmt),
                                       pmt.init_u8vector(len(packet), packet)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks


class hexdump_sink(gr.hier_block2):
    """
    Hierarchical block for hexdump output

    The input are PDUs with frames.

    These are printed in hex.

    Args:
        options: options from argparse
    """
    def __init__(self, options=None):
        gr.hier_block2.__init__(
            self,
            'hexdump_sink',
            gr.io_signature(0, 0, 0),
            gr.io_signature(0, 0, 0))
        self.message_port_register_hier_in('in')

        self.message_debug = blocks.message_debug()
        self.msg_connect((self, 'in'), (self.message_debug, 'print_pdu'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#


from gnuradio import gr, digital, blocks, fec

from ... import descrambler308
from ... import k2sat_deframer as deframer
from ...hier.sync_to_pdu_packed import sync_to_pdu_packed
from ...utils.options_block import options_block


_syncword = '0101010101111110'


class k2sat_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe the K2SAT custom image protocol

    The input is a float stream of soft symbols. The output are PDUs
    with frames.

    Args:
        syncword_threshold: number of bit errors allowed in syncword (int)
        options: Options from argparse
    """
    def __init__(self, syncword_threshold=None, options=None):
        gr.hier_block2.__init__(
            self,
            'k2sat_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        if syncword_threshold is None:
            syncword_threshold = self.options.syncword_threshold

        # For QPKS 90º abiguity
        self.deinterleave = blocks.deinterleave(gr.sizeof_float, 1)
        self.invert = blocks.multiply_const_ff(-1, 1)
        self.interleave = blocks.interleave(gr.sizeof_float, 1)

        self.cc0 = fec.cc_decoder.make(
            80, 7, 2, [79, 109], 0, -1, fec.CC_STREAMING, False)
        self.cc1 = fec.cc_decoder.make(
            80, 7, 2, [79, 109], 0, -1, fec.CC_STREAMING, False)
        self.viterbi0 = fec.extended_decoder(
            decoder_obj_list=self.cc0, threading=None, ann=None,
            puncpat='11')
        self.viterbi1 = fec.extended_decoder(
            decoder_obj_list=self.cc1, threading=None, ann=None,
            puncpat='11')
        self.diff0 = digital.diff_decoder_bb(2)
        self.diff1 = digital.diff_decoder_bb(2)
        self.scrambler0 = descrambler308()
        self.scrambler1 = descrambler308()
        self.deframer0 = sync_to_pdu_packed(
            packlen=2200, sync=_syncword, threshold=syncword_threshold)
        self.deframer1 = sync_to_pdu_packed(
            packlen=2200, sync=_syncword, threshold=syncword_threshold)
        self.deframer = deframer()

        self.connect(self, self.viterbi0, self.diff0,
                     self.scrambler0, self.deframer0)
        self.connect(self, self.deinterleave)
        self.connect((self.deinterleave, 0), (self.interleave, 1))
        self.connect((self.deinterleave, 1), self.invert, (self.interleave, 0))
        self.connect(self.interleave, self.viterbi1, self.diff1,
                     self.scrambler1, self.deframer1)

        self.msg_connect((self.deframer0, 'out'), (self.deframer, 'in'))
        self.msg_connect((self.deframer1, 'out'), (self.deframer, 'in'))
        self.msg_connect((self.deframer, 'out'), (self, 'out'))

    _default_sync_threshold = 0

    @classmethod
    def add_options(cls, parser):
        """
        Adds K2SAT deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--syncword_threshold', type=int,
            default=cls._default_sync_threshold,
            help='Syncword bit errors [default=%(default)r]')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import subprocess


class FehOpener():
    def __init__(self, fullscreen=True, interval=1):
        self.fullscreen = fullscreen
        self.interval = interval

    def __params(self):
        p = ['-F'] if self.fullscreen else []
        return p + ['-R', str(self.interval)]

    def open(self, filename):
        subprocess.Popen(['feh'] + self.__params() + [filename],
                         stdin=subprocess.DEVNULL)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import numpy


flag = bytes([0] + 6*[1] + [0])


def crc_ccitt(data):
    # Implementation taken from
    # hdlc_framer_pb_impl.cc in gr-digital
    poly = 0x8408  # reflected 0x1021
    crc = 0xffff
    for byte in data:
        crc ^= byte
        for _ in range(8):
            crc = (crc >> 1) ^ poly if crc & 1 else crc >> 1
    return crc ^ 0xffff



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr
import pmt


# Details of the NGHam protocol taken from
# https://github.com/skagmo/ngham/blob/master/ngham.c

ngham_rs_sizes = [47, 79, 111, 159, 191, 223, 255]
ngham_non_rs_sizes = [47-16, 79-16, 111-16, 159-32, 191-32, 223-32, 255-32]


class ngham_remove_padding(gr.basic_block):
    """
    Removes the padding on an NGHam packet

    Also it automatically detects if RS check bytes are
    still in the packet and removes them as well
    """
    def __init__(self):
        gr.basic_block.__init__(
            self,
            name='ngham_remove_padding',
            in_sig=[],
            out_sig=[])
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = bytes(pmt.u8vector_elements(msg))

        if len(packet) in ngham_rs_sizes:
            size_idx = ngham_rs_sizes.index(len(packet))
            packet = packet[:ngham_non_rs_sizes[size_idx]]

        if len(packet) not in ngham_non_rs_sizes:
            print('NGHam packet length invalid')
            return

        padding = packet[0] & 0x1f
        packet = packet[:-padding]
        self.message_port_pub(
            pmt.intern('out'),
            pmt.cons(pmt.car(msg_pmt), pmt.init_u8vector(len(packet), packet)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from math import log10

from construct import *

from .csp import CSPHeader
from ..adapters import LinearAdapter


PayloadMode = BitStruct(
    'open_telecommand' / Flag,
    'camera_task' / Flag,
    'valid_image_data' / Flag,
    'camera_power' / Flag,
    'rest' / Nibble
    )

Iv3 = LinearAdapter(10.0, Int16sl)
Ivbatt = LinearAdapter(2.5, Int16sl)
Voltage = LinearAdapter(2000.0, Int16sl)


class TSTM32Adapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int(4096.0*(2.5*(obj - 25.0) + 760.0)/3000.0)

    def _decode(self, obj, context, path=None):
        return (obj * 3000.0 / 4096.0 - 760.0) / 2.5 + 25.0


TSTM32 = TSTM32Adapter(Int16sl)


class TPAAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int(obj*16.0) << 3

    def _decode(self, obj, context, path=None):
        return (obj >> 3) / 16.0


TPA = TPAAdapter(Int16sl)


DCFMTC = LinearAdapter(352.7, Int16sl)
DCFMHam = LinearAdapter(6864.0, Int16sl)


class RSSIAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int(10**((obj + 147.0)/10.0))

    def _decode(self, obj, context, path=None):
        return 10*log10(obj) - 147.0


RSSI = RSSIAdapter(Int32ul)

Runtime = LinearAdapter(1000.0, Int32ul)
CTCSSCount = LinearAdapter(0.04, Int32ul)

# COM STM32 Housekeeping Packet
Hk_STM32 = Struct(
    Const(b'\x1c\xa1'),
    'config' / Int8ul,
    'flag_direct_ins' / Int8ul,
    'payload_mode' / PayloadMode,
    'tx_mode' / Int8ul,
    'gain_tx' / Int16sl,
    'i_3v3' / Iv3,
    'u_3v3' / Voltage,
    'i_vbat_tx' / Ivbatt,
    'u_vbat_tx' / Voltage,
    'i_vbat_rx' / Ivbatt,
    'u_vbat_rx' / Voltage,
    't_stm32' / TSTM32,
    't_pa' / TPA,
    'n_tx_rf' / Int16ul,
    'n_rx_rf' / Int16ul,
    'n_tx_err_rf' / Int16ul,
    'n_tx_err_rf' / Int16ul,
    'n_tx_i2c' / Int16ul,
    'n_rx_i2c' / Int16ul,
    'n_tx_err_i2c' / Int16ul,
    'n_rx_err_i2c' / Int16ul,
    'n_tc' / Int32ul,
    'dc_fm_tc' / DCFMTC,
    'dc_fm_ham' / DCFMHam,
    'rssi_fm_tc' / RSSI,
    'rssi_fm_ham' / RSSI,
    'reset_flag' / Int8ul,
    'sys_flag' / Int8ul,
    'dma_overflow' / Int16ul,
    'runtime' / Runtime,
    'reset_count' / Int32ul,
    'ctcss_count' / CTCSSCount,
    'ctcss_det' / Float32l
    )

# COM STM32 Config Packet
Cfg = Struct(
    Const(b'\x1c\xa2'),
    'gain_tx_HI' / Int16sl,
    'gain_tx_LO' / Int16sl,
    'bias_I' / Int16sl,
    'bias_Q' / Int16sl,
    'threshold_u_vbat_rx_powerdown' / Int16sl,
    'threshold_u_vbat_rx_repeateroff' / Int16sl,
    'threshold_t_pa' / Int8sl,
    'cam_ham_interval' / Int8ul,
    'cam_ham_en' / Int8ul,
    'ctcss_en' / Int8ul,
    'ctcss_n_integration' / Int8ul,
    'ctcss_n_tail' / Int8ul,
    'ctcss_coeff' / Float32l,
    'ctcss_threshold' / Float32l,
    'gain_fmdm_ham' / Float32l,
    'gain_fmdm_tc' / Float32l,
    'interval_hk_OBC' / Int32ul,
    'interval_hk_TLM' / Int32ul,
    'interval_hk_BEACON' / Int32ul,
    'message' / Bytes(28),
    'cam_delay' / Int32ul,
    'crc' / Int32ul)

# COM AVR Housekeeping Packet
Hk_AVR = Struct(
    Const(b'\x1c\xa3'),
    'adf7021_ld' / Int8ul,
    'err_flag' / Int8ul,
    'n_tx_i2c' / Int16ul,
    'n_rx_i2c' / Int16ul,
    'n_tx_232' / Int16ul,
    'n_rx_232' / Int16ul,
    'runtime' / Runtime,
    'rssi_analog' / Int8ul,
    'n_rssi_const' / Int8ul,
    'unlock_count' / Int8ul,
    'reset_flag' / Int8ul,
    'reset_count' / Int32ul)

# LilacSat-1 Generic Telemetry Data
LilacSat1Tlm = Struct(
    Const(b'\xaa\xa1'),
    'ttc_payload_mode' / Int8ul,
    'ttc_tx_mode' / Int8ul,
    'ttc_i_vbat_tx' / Int8ul,
    'ttc_t_pa' / Int8sl,
    'ttc_adf7021_lock' / Int8ul,
    'ttc_n_tc' / Int32ul,
    'ttc_rssi_fm_tc' / Int8ul,
    'eps_bat_vol_average' / Int8ul,
    'eps_bat_cur' / Int8sl,
    'eps_5v_cur' / Int8ul,
    'esp_3v3_cur' / Int8ul,
    'eps_5v_dcdc_temp' / Int16sl,
    'eps_bat_temp' / Int16sl,
    'eps_reserved' / Byte[2],
    'adcs_q1' / Int16sl,
    'adcs_q2' / Int16sl,
    'adcs_q3' / Int16sl,
    'adcs_w1' / Int16sl,
    'adcs_w2' / Int16sl,
    'adcs_w3' / Int16sl,
    'adcs_mode' / Int16sl,
    'adcs_long' / Int16sl,
    'adcs_lat' / Int16sl,
    'adcs_height' / Int16sl,
    'adcs_reserved' / Byte[5],
    'obc_gps_power_state' / Int8ul,
    'obc_eps_low_state' / Int8ul,
    'obc_year' / Int16ul,
    'obc_month' / Int8ul,
    'obc_day' / Int8ul,
    'obc_hour' / Int8ul,
    'obc_minute' / Int8ul,
    'obc_second' / Int8ul,
    'obc_reserved' / Byte,
    'inms_script_en' / Int8ul,
    'inms_power_on' / Int8ul,
    'inms_script_no' / Int8ul,
    'inms_script_seq' / Int8ul,
    'inms_script_execed' / Int8ul,
    'inms_script_checksum_ok' / Int8ul,
    'inms_slot_not_empty' / Int8ul,
    'inms_reserved' / Byte[2])

# OBC Housekeeping Parameters
Hk_OBC = Struct(
    Const(b'\x1a\xa1'),
    'com_iic1_bus_status' / Int8ul,
    'eps_iic1_bus_status' / Int8ul,
    'gps_power_status' / Int8ul,
    'battery_low_indicator' / Int8ul,
    'gps_lock_status' / Int8ul,
    'onboard_time_source' / Int8ul,
    'time_onboard_year' / Int16ul,
    'time_onboard_month' / Int8ul,
    'time_onboard_day' / Int8ul,
    'time_onboard_hour' / Int8ul,
    'time_onboard_min' / Int8ul,
    'time_onboard_sec' / Int8ul,
    'delayed_command_count' / Int8ul,
    'interface_board_3v3_current_a' / Int16ul,
    'interface_board_5v_current_b' / Int16ul,
    'obc_board_3v3_current_tsc103' / Int16ul,
    'obc_board_5v_current_ina219' / Int16ul,
    'obc_board_5v_voltage_ina219' / Int16ul,
    Padding(2),
    'csp_sent_packet' / Int32ul,
    'csp_received_packet' / Int32ul,
    'received_command_count' / Int32ul,
    'obc_power_on_time' / Int32ul,
    'obc_reset_count' / Int16ul,
    'obc_log_file_index' / Int16ul,
    'com_log_file_index' / Int16ul,
    'eps_log_file_index' / Int16ul,
    'com_log_2_file_index' / Int16ul,
    'inms_enable_status_script_power' / Int8ul,
    'inms_running_status_script_no_seq_no' / Int8ul,
    'adcs_log_file_index' / Int16ul,
    'inms_temperature_sensing_point' / Float32l,
    'stm32_temperature' / Int16ul,
    'current_storage_card_telemetry_ftp' / Int8ul,
    'gps_leap_second_correction' / Int8sl,
    'inms_file_index' / Int32ul)

Magnetometer = LinearAdapter(1/4.0, Int16sl)
Gyroscope = LinearAdapter(32768.0/57.2958, Int16sl)
TempADIS = LinearAdapter(1/0.0739, Int16sl)
FineSun = LinearAdapter(16384.0, Int16sl)
CoarseSun = LinearAdapter(32.0, Int16sl)
Latitude = LinearAdapter(360.0, Int16sl)
Longitude = LinearAdapter(180.0, Int16sl)
Altitude = LinearAdapter(1/20.0, Int16sl)

# ADCS Parameters (OBC)
ADCS_Par = Struct(
    Const(b'\x1a\xa2'),
    'magnetometer_hmc5883l' / Magnetometer[3],
    'magnetometer_hmc1043l' / Magnetometer[3],
    'gyroscope_mpu3300' / Gyroscope[3],
    'gyroscope_mpu3300_temperature' / Int16sl,
    'gyroscope_adis16448' / Gyroscope[3],
    'temperature_adis16448' / TempADIS,
    'magnetometer_adis16448' / Magnetometer[3],
    'fine_sun_sensor' / FineSun[3],
    'coarse_sun_sensor' / CoarseSun[5],
    'magnetometer_mix' / Float32l,
    'modeflag_2' / Int32ul,
    'modeflag_1' / Int16ul,
    'flywheel_speed' / Int16sl,
    'adcs_interruption_counting' / Int16ul,
    'kf_quaternions' / FineSun[3],
    'gyroscope_quaternions' / FineSun[3],
    'coarse_sun_sensor_temperature' / Int16sl[5],
    'latitude' / Latitude,
    'longitude' / Longitude,
    'altitude' / Altitude,
    'fine_sun_sensor_temperature' / Int16sl)

# Orbit Paramaters (OBC)
Orbit_Par = Struct(
    Const(b'\x1a\xa8'),
    'year' / Int8ul,
    'month' / Int8ul,
    'day' / Int8ul,
    'hour' / Int8ul,
    'minute' / Int8ul,
    'second' / Int8ul,
    'position' / Float64l[3],
    'velocity' / Float64l[3],
    'oev_a' / Float64l,
    'oev_e' / Float64l,
    'oev_i' / Float64l,
    'oev_omega' / Float64l,
    'oev_w' / Float64l,
    'oev_theta' / Float64l)

# OBC Variable Threshold Parameter
OBC_Threshold = Struct(
    Const(b'\x1a\xa9'),
    'eps_batter_low_voltage_threshold' / Int16ul,
    'eps_overcurrent_threshold' / Int16ul,
    'eps_critical_battery_low_voltage_threshold' / Int16ul,
    'eps_critical_overcurrent_threshold' / Int16ul,
    'eps_overcurrent_hysteresis' / Int16ul,
    'gps_time_error_threshold' / Int16ul,
    'gps_position_malfunction_threshold' / Int16ul[3],
    'gps_velocity_malfunction_threshold' / Int16ul[3])

# Orbit Parameter (OBC)
Orbital_Par = Struct(
    Const(b'\x1a\xaa'),
    'packet_header' / Int16ul,
    'frame_header' / Int8ul,
    'frame_length' / Int8ul,
    'gps_bd_lock_status' / Int8ul,
    'week_of_year' / Int16ul,
    'second_of_week' / Float64l,
    'gps_satellite_in_use_obs' / Int8ul,
    'gps_satellite_in_use_pdop' / Int16ul,
    'xyz' / Float64l[3],
    'v' / Float64l[3],
    'gps_clock_error' / Float32l,
    'gps_clock_drift' / Float32l,
    'bd_clock_error' / Float32l,
    'bd_clock_drift' / Float32l,
    'checksum' / Int8ul)

CurrentVoltage = LinearAdapter(1000.0, Int16sl)
Temp = LinearAdapter(16.0, Int16sl)

EPS_Status = BitStruct(
    'reboot_enable_2_status' / Flag,
    Padding(1),
    'reboot_enable_1_status' / Flag,
    Padding(1),
    'antenna_2_deploy_status' / Flag,
    Padding(1),
    'antenna_1_deploy_status' / Flag,
    Padding(1),
    'antenna_deploy_disable_control' / Flag,
    Padding(7))

# EPS Housekeeping Parameter
Hk_EPS = Struct(
    Const(b'\x1e\xa1'),
    'mppt_output_current' / CurrentVoltage[5],
    'battery_output_current' / CurrentVoltage[3],
    '5v_output_current' / CurrentVoltage,
    '3v3_output_current' / CurrentVoltage,
    'main_bus_output_current' / CurrentVoltage,
    'antenna_deployment_current' / CurrentVoltage,
    'mppt_output_voltage' / CurrentVoltage[5],
    'battery_output_voltage' / CurrentVoltage[3],
    '5v_output_voltage' / CurrentVoltage,
    '3v3_output_voltage' / CurrentVoltage,
    'main_bus_output_voltage' / CurrentVoltage,
    'antenna_deployment_voltage' / CurrentVoltage,
    'mptt_temperature' / Temp,
    'battery_charger_temperature' / Temp,
    'battery_temperature' / Temp,
    'main_bus_diode_temperature' / Temp,
    '5v_regulator_temperature' / Temp,
    '3v3_regulator_temperature' / Temp,
    '5v_ocp_temperature' / Temp,
    '3v3_ocp_temperature' / Temp,
    'status' / EPS_Status,
    'power_on_time' / LinearAdapter(1000.0, Int32ul),
    'iic_sent_packet_count' / Int16ul,
    'iic_received_packet_count' / Int16ul,
    'output_reboot_wdt_in_minute' / Int16ul,
    'eps_reboot_wdt_in_minute' / Int16ul,
    'eps_reboot_count' / Int16ul)

# EPS Onboard Parameters
EPS_Onboard = Struct(
    Const(b'\x1e\xa2'),
    'onboard_software_version' / Int8ul[3],
    'onboard_software_compile_date' / Int8ul[4],
    Padding(1),
    'first_poweron_flag' / Int8ul,
    Padding(3),
    'output_reboot_wdt_enabled' / Int8ul,
    'output_reboot_timer_hr' / Int8ul,
    Padding(2),
    'eps_reboot_wdt_enabled' / Int8ul,
    'eps_reboot_timer_hr' / Int8ul,
    Padding(2),
    'auto_send_hk_data_enabled' / Int8ul,
    'auto_send_hk_data_period_sec' / Int8ul,
    Padding(2),
    'antenna_deployed' / Int8ul)

# EPS Reboot Log
EPS_Reboot = Struct(
    Const(b'\x1e\xa3'),
    'onboard_software_version' / Int8ul[3],
    'onboard_software_compile_date' / Int8ul[4],
    Padding(1),
    'last_output_reboot_source' / Int8ul,
    'last_output_reboot_timer' / Int8ul[4],
    Padding(3),
    'last_eps_reboot_source' / Int8ul,
    'last_eps_reboot_timer' / Int8ul[4],
    Padding(3),
    'current_onboard_reboot_timer' / Int8ul[4],
    Padding(4),
    'eps_reboot_count' / Int8ul[4])

# EPS Command Log
Command = Struct(
    'cmd' / Int8ul[5],
    'source' / Int8ul,
    'received_time' / Int8ul[4])


EPS_Command = Struct(
    Const(b'\x1e\xa4'),
    'cmds' / Command[4])

Error = Struct(
    'code' / Int8ul,
    'time' / Int8ul[4])

# EPS Error Log
EPS_Error = Struct(
    Const(b'\x1e\xa5'),
    'critical_error' / Error[4],
    'error' / Error[4])

ImageData = Struct(
    'file_id' / Int32ul,
    'flag' / Int8sb,
    'filesize' / Int24ul,
    'offset' / Int24ul,
    'data' / Bytes(64)
    )

Frame = Select(
    Hk_STM32, Cfg, Hk_AVR, LilacSat1Tlm, Hk_OBC, ADCS_Par, Orbit_Par,
    OBC_Threshold, Orbital_Par, Hk_EPS, EPS_Onboard, EPS_Reboot, EPS_Command,
    EPS_Error)

by70_1 = Struct(
    'csp_header' / CSPHeader,
    'beacon' / If(this.csp_header.destination == 5, Frame),
    'camera' / If(this.csp_header.destination == 6, ImageData),
    'packet_count' / Int32ul,
    'crc' / Hex(Int32ul)
    )

taurus1 = Struct(
    'header' / Bytes(5+7),  # This is a CCSDS header of some sort
    'beacon' / Frame
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 @dbusan <https://github.com/dbusan>
# Copyright 2021 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *

from .ax25 import Header as AX25Header


telem_params = {
    # highR, lowR, offset, scale
    'V_3V3': (10000, 10000, 0, 1),
    'V_Batt': (24500, 10000, 0, 1),
    'V_SolarPanelXX': (25000, 10000, 0, 1),
    'V_SolarPanelYY': (25000, 10000, 0, 1),
    'I_3V3': (0, 10000, 0.25, 0.8),
    'I_Batt': (0, 10000, 0.25, 0.8),
    'I_SolOutXX': (0, 10000, 0.25, 0.8),
    'I_SolInXX': (0, 10000, 0.25, 0.8),
    'I_SolOutYY': (0, 10000, 0.25, 0.8),
    'I_SolInYY': (0, 10000, 0.25, 0.8),
    'I_Magnetorquers0XX': (0, 10000, 0, 0.25),
    'I_Magnetorquers0YY': (0, 10000, 0, 0.25),
    'I_Magnetorquers0ZZ': (0, 10000, 0, 0.25),
    'I_Magnetorquers1XX': (0, 10000, 0, 0.25),
    'I_Magnetorquers1YY': (0, 10000, 0, 0.25),
    'T_SolNegXX': (0, 10000, 67.8 / 160.0, 1.0 / 160.0),
    'T_SolPosXX': (0, 10000, 67.8 / 160.0, 1.0 / 160.0),
    'T_SolNegYY': (0, 10000, 67.8 / 160.0, 1.0 / 160.0),
    'T_SolPosYY': (0, 10000, 67.8 / 160.0, 1.0 / 160.0),
    'T_BHeat12': (0, 10000, 50.0 / 100.0, 1.0 / 100.0),
    'T_BHeat34': (0, 10000, 50.0 / 100.0, 1.0 / 100.0),
}


class BinarAdapter(Adapter):
    def __init__(self, telemetry_id, *args, **kwargs):
        self.telemetry_id = telemetry_id
        return Adapter.__init__(self, *args, **kwargs)

    def _encode(self, obj, context, path=None):
        pass

    def _decode(self, obj, context, path=None):
        telem_obj_params = telem_params[self.telemetry_id]

        highR = telem_obj_params[0]
        lowR = telem_obj_params[1]
        offsetvalue = telem_obj_params[2]
        scaleratio = telem_obj_params[3]

        return generalised_decode(obj, lowR, highR, scaleratio, offsetvalue)


Voltages = Struct(
    'V_3V3' / BinarAdapter('V_3V3', Int16ul),
    'V_Batt' / BinarAdapter('V_Batt', Int16ul),
    'V_SolarPanelXX' / BinarAdapter('V_SolarPanelXX', Int16ul),
    'V_SolarPanelYY' / BinarAdapter('V_SolarPanelYY', Int16ul),
)

Currents = Struct(
    'I_3V3' / BinarAdapter('I_3V3', Int16ul),
    'I_Batt' / BinarAdapter('I_Batt', Int16ul),
    'I_SolOutXX' / BinarAdapter('I_SolOutXX', Int16ul),
    'I_SolInXX' / BinarAdapter('I_SolInXX', Int16ul),
    'I_SolOutYY' / BinarAdapter('I_SolOutYY', Int16ul),
    'I_SolInYY' / BinarAdapter('I_SolInYY', Int16ul),
    'I_Magnetorquers0XX' / BinarAdapter('I_Magnetorquers0XX', Int16ul),
    'I_Magnetorquers0YY' / BinarAdapter('I_Magnetorquers0YY', Int16ul),
    'I_Magnetorquers0ZZ' / BinarAdapter('I_Magnetorquers0ZZ', Int16ul),
    'I_Magnetorquers1XX' / BinarAdapter('I_Magnetorquers1XX', Int16ul),
    'I_Magnetorquers1YY' / BinarAdapter('I_Magnetorquers1YY', Int16ul),
)

Temperatures = Struct(
    'T_SolNegXX' / BinarAdapter('T_SolNegXX', Int16ul),
    'T_SolPosXX' / BinarAdapter('T_SolPosXX', Int16ul),
    'T_SolNegYY' / BinarAdapter('T_SolNegYY', Int16ul),
    'T_SolPosYY' / BinarAdapter('T_SolPosYY', Int16ul),
    'T_BHeat12' / BinarAdapter('T_BHeat12', Int16ul),
    'T_BHeat34' / BinarAdapter('T_BHeat34', Int16ul),
)

Telemetry = Struct(
    'Voltages' / Voltages,
    'Currents' / Currents,
    'Temperatures' / Temperatures,
    'GPSPosn' / Float32l[2],
)

StringMsg = Struct(
    'Message' / Bytes(36)
)


def generalised_decode(obj, lowR, highR, scaleratio, offsetvalue):
    vref = 3.36
    vdiv_ratio = float(lowR)/float(highR+lowR)

    val = (obj/2**16*vref/vdiv_ratio-offsetvalue)/(scaleratio)
    return val


binar1 = Struct(
    'packetlen' / Int8ub,
    'telemetry' / Telemetry,
    'message' / StringMsg,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, digital, fec, blocks

from ... import (
    distributed_syncframe_soft, matrix_deinterleaver_soft, decode_rs,
    check_tt64_crc,
    )
from ...hier.ccsds_descrambler import ccsds_descrambler
from ...utils.options_block import options_block


_syncword = '11111110000111011110010110010010000001000100110001011101011011000'
_syncword_short = '1111111000011101111001011001001000000100010011000101'


class ao40_fec_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe the AO-40 FEC protocol.

    The input is a float stream of soft symbols. The output are PDUs
    with frames.

    Args:
        syncword_threshold: number of bit errors allowed in syncword (int)
        short_frames: use short frames (used in SMOG-P) (bool)
        crc: use CRC-16 ARC (used in SMOG-P) (bool)
        options: Options from argparse
    """
    def __init__(self, syncword_threshold=None, short_frames=False,
                 crc=False, options=None):
        gr.hier_block2.__init__(
            self,
            'ao40_fec_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        if syncword_threshold is None:
            syncword_threshold = self.options.syncword_threshold

        self.deframer = distributed_syncframe_soft(
            syncword_threshold, _syncword_short if short_frames else _syncword,
            51 if short_frames else 80)
        self.deinterleaver = matrix_deinterleaver_soft(
            51 if short_frames else 80, 52 if short_frames else 65,
            2572 if short_frames else 5132, 80 if short_frames else 65)
        self.viterbi = fec.cc_decoder.make(
            2572 if short_frames else 5132, 7, 2, [79, -109], 0, -1,
            fec.CC_TERMINATED, False)
        self.viterbi_decoder = fec.async_decoder(
            self.viterbi, False, False, 2572//8 if short_frames else 5132//8)
        self.scrambler = ccsds_descrambler()
        self.rs = decode_rs(False, 1 if short_frames else 2)

        if crc:
            # CRC-16 ARC
            self.crc = check_tt64_crc(
                verbose=self.options.verbose_crc, packet_len=None,
                strip=False)

        self.connect(self, self.deframer)
        self.msg_connect((self.deframer, 'out'), (self.deinterleaver, 'in'))
        self.msg_connect((self.deinterleaver, 'out'),
                         (self.viterbi_decoder, 'in'))
        self.msg_connect((self.viterbi_decoder, 'out'), (self.scrambler, 'in'))
        self.msg_connect((self.scrambler, 'out'), (self.rs, 'in'))
        if crc:
            self.msg_connect((self.rs, 'out'), (self.crc, 'in'))
            self.msg_connect((self.crc, 'ok'), (self, 'out'))
        else:
            self.msg_connect((self.rs, 'out'), (self, 'out'))

    _default_sync_threshold = 8

    @classmethod
    def add_options(cls, parser):
        """
        Adds AO-40 FEC deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--syncword_threshold', type=int,
            default=cls._default_sync_threshold,
            help='Syncword bit errors [default=%(default)r]')
        parser.add_argument(
            '--verbose_rs', action='store_true', help='Verbose RS decoder')
        parser.add_argument(
            '--verbose_crc', action='store_true', help='Verbose CRC decoder')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from math import ceil, pi
import pathlib
import sys

from gnuradio import gr, blocks, analog, digital, filter
from gnuradio.filter import firdes
import numpy as np

from ...hier.rms_agc_f import rms_agc_f
from ...utils.options_block import options_block


class fsk_demodulator(gr.hier_block2, options_block):
    """
    Hierarchical block to demodulate FSK.

    The input can be either IQ or real. For IQ input, it is assumed
    that the data is FM modulated, so FM demodulation is performed.
    For real input, it is assumed that the data is already FM
    demodulated.

    Args:
        baudrate: Baudrate in symbols per second (float)
        sample_rate: Sample rate in samples per second (float)
        iq: Whether the input is IQ or real (bool)
        deviation: Deviation in Hz, negative inverts sidebands (float)
        subaudio: Use subaudio demodulation (bool)
        dc_block: Use DC-block (bool)
        dump_path: Path to dump internal signals to files (str)
        options: Options from argparse
    """
    def __init__(self, baudrate, samp_rate, iq, deviation=None,
                 subaudio=False, dc_block=True, dump_path=None,
                 options=None):
        gr.hier_block2.__init__(
            self,
            'fsk_demodulator',
            gr.io_signature(1, 1,
                            gr.sizeof_gr_complex if iq else gr.sizeof_float),
            gr.io_signature(1, 1, gr.sizeof_float))
        options_block.__init__(self, options)

        use_agc = self.options.use_agc or not iq
        if self.options.disable_dc_block:
            dc_block = False

        if dump_path is not None:
            dump_path = pathlib.Path(dump_path)

        if deviation is not None:
            _deviation = deviation
        else:
            _deviation = self.options.deviation

        # Prevent problems due to baudrate too high
        if baudrate >= samp_rate:
            print(f'Sample rate {samp_rate} sps insufficient for {baudrate} '
                  'baud FSK demodulation. Demodulator will not work.',
                  file=sys.stderr)
            baudrate = samp_rate / 2

        if iq:
            # Cut to Carson's bandwidth rule before quadrature demod.
            # Note that _deviation can be negative to encode that the
            # low tone corresponds to the symbol 1 and the high tone
            # corresponds to the symbol 0.
            carson_cutoff = abs(_deviation) + baudrate / 2
            self.demod = analog.quadrature_demod_cf(
                samp_rate/(2*pi*_deviation))
            if carson_cutoff >= samp_rate / 2:
                # Sample rate is already narrower than Carson's
                # bandwidth. Do not filter
                self.connect(self, self.demod)
            else:
                # Sample rate is wider than Carson's bandwidth.
                # Lowpass filter before demod.
                fir_taps = firdes.low_pass(
                    1, samp_rate, carson_cutoff, 0.1 * carson_cutoff)
                self.demod_filter = filter.fir_filter_ccf(1, fir_taps)
                self.connect(self, self.demod_filter, self.demod)
        else:
            self.demod = self

        sps = samp_rate / baudrate
        max_sps = 10
        if sps > max_sps:
            decimation = ceil(sps / max_sps)
        else:
            decimation = 1
        sps /= decimation

        if subaudio:
            # some not-so-bad filter parameters for subaudio processing
            subaudio_cutoff = 2.0/3.0 * baudrate
            subaudio_transition = subaudio_cutoff / 4.0
            subaudio_taps = firdes.low_pass(
                1, samp_rate, subaudio_cutoff, subaudio_transition)
            self.subaudio_lowpass = filter.fir_filter_fff(1, subaudio_taps)

        # square pulse filter
        sqfilter_len = int(samp_rate / baudrate)
        taps = np.ones(sqfilter_len)/sqfilter_len
        self.lowpass = filter.fir_filter_fff(decimation, taps)

        if dc_block:
            self.dcblock = filter.dc_blocker_ff(ceil(sps * 32), True)
        else:
            self.dcblock = self.lowpass  # to simplify connections below

        if use_agc:
            # This gives a time constant of 50 symbols
            agc_constant = 2e-2 / sps
            self.agc = rms_agc_f(agc_constant, 1)

        if dump_path is not None:
            self.waveform = blocks.file_sink(
                gr.sizeof_float, str(dump_path / 'waveform.f32'))

        # "Eempiric" formula for TED gain of Gardner detector
        # 1.47 symbol^{-1}
        ted_gain = 1.47
        damping = 1.0
        self.clock_recovery = digital.symbol_sync_ff(
            digital.TED_GARDNER, sps, self.options.clk_bw, damping, ted_gain,
            self.options.clk_limit * sps, 1,
            digital.constellation_bpsk().base(), digital.IR_PFB_NO_MF)

        if dump_path is not None:
            self.clock_recovery_out = blocks.file_sink(
                gr.sizeof_float,
                str(dump_path / 'clock_recovery_out.f32'), False)
            self.clock_recovery_err = blocks.file_sink(
                gr.sizeof_float,
                str(dump_path / 'clock_recovery_err.f32'), False)
            self.clock_recovery_T_inst = blocks.file_sink(
                gr.sizeof_float,
                str(dump_path / 'clock_recovery_T_inst.f32'), False)
            self.clock_recovery_T_avg = blocks.file_sink(
                gr.sizeof_float,
                str(dump_path / 'clock_recovery_T_avg.f32'), False)
            self.connect(self.clock_recovery, self.clock_recovery_out)
            self.connect((self.clock_recovery, 1), self.clock_recovery_err)
            self.connect((self.clock_recovery, 2), self.clock_recovery_T_inst)
            self.connect((self.clock_recovery, 3), self.clock_recovery_T_avg)

        conns = [self.demod]
        if subaudio:
            conns.append(self.subaudio_lowpass)
        conns.append(self.lowpass)
        if dc_block:
            conns.append(self.dcblock)
        self.connect(*conns)
        if use_agc:
            self.connect(self.dcblock, self.agc, self.clock_recovery)
            if dump_path is not None:
                self.connect(self.agc, self.waveform)
        else:
            self.connect(self.dcblock, self.clock_recovery)
            if dump_path is not None:
                self.connect(self.dcblock, self.waveform)
        self.connect(self.clock_recovery, self)

    _default_clk_rel_bw = 0.06
    _default_clk_limit = 0.004
    _default_deviation_hz = 5000

    @classmethod
    def add_options(cls, parser):
        """
        Adds FSK demodulator specific options to the argparse parser
        """
        parser.add_argument(
            '--clk_bw', type=float, default=cls._default_clk_rel_bw,
            help=('Clock recovery bandwidth (relative to baudrate) '
                  '[default=%(default)r]'))
        parser.add_argument(
            '--clk_limit', type=float, default=cls._default_clk_limit,
            help=('Clock recovery limit (relative to baudrate) '
                  '[default=%(default)r]'))
        parser.add_argument(
            '--deviation', type=float, default=cls._default_deviation_hz,
            help='Deviation (Hz) [default=%(default)r]')
        parser.add_argument(
            '--use_agc', action='store_true',
            help='Use AGC (for IQ input. AGC always on for real input)')
        parser.add_argument(
            '--disable_dc_block', action='store_true',
            help='Disable DC block')



#!D:\bld\gnuradio-satellites_1642173413548\_h_env\python.exe
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""Plots SMOG-P/ATL-1 spectrum data previously saved to a file"""

import pathlib
import re
import sys

import matplotlib.pyplot as plt
import numpy as np


def usage():
    print(f'Usage: {sys.argv[0]} input_file', file=sys.stderr)
    sys.exit(1)


def main():
    if len(sys.argv) != 2:
        usage()

    filename = sys.argv[1]
    m = re.match(
        'spectrum_start_(?P<start>\\d+?)_step_(?P<step>\\d+?)_'
        'rbw_\\d+_measid_(?P<measid>\\d+?)\\Z',
        pathlib.Path(filename).name)
    start = int(m.group('start'))
    step = int(m.group('step'))
    measid = int(m.group('measid'))

    data = np.fromfile(filename, dtype='uint8')
    freq = start + step * np.arange(data.size)

    plt.figure()
    plt.plot(freq * 1e-6, data)
    plt.title(f'Spectrum measurement {measid}')
    plt.xlabel('Frequency (MHz)')
    plt.ylabel('PSD')
    plt.savefig(pathlib.Path(filename).parent / f'spectrum_{measid}.png')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, digital

from ... import decode_rs
from ...hier.sync_to_pdu import sync_to_pdu
from ...hier.sync_to_pdu_packed import sync_to_pdu_packed
from ...hier.ccsds_descrambler import ccsds_descrambler
from ...utils.options_block import options_block


_syncword = '00011010110011111111110000011101'


class ccsds_rs_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe CCSDS Reed-Solomon TM frames

    The input is a float stream of soft symbols. The output are PDUs
    with frames.

    Args:
        frame_size: frame size (not including parity check bytes) (int)
        precoding: either None or 'differential' for differential precoding
                   (str)
        rs_en: If Reed-Solomon should be enabled or not (bool)
        rs_basis: Reed-Solomon basis, either 'conventional' or 'dual' (str)
        rs_interleaving: Reed-Solomon interleaving depth (int)
        scrambler: scrambler to use, either 'CCSDS' or 'none' (str)
        syncword_threshold: number of bit errors allowed in syncword (int)
        options: Options from argparse
    """
    def __init__(self, frame_size=223, precoding=None, rs_en=True,
                 rs_basis='dual', rs_interleaving=1, scrambler='CCSDS',
                 syncword_threshold=None, options=None):
        gr.hier_block2.__init__(
            self,
            'ccsds_rs_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        if precoding not in [None, 'differential']:
            raise ValueError(f'invalid precoding {precoding}')
        if rs_basis not in ['conventional', 'dual']:
            raise ValueError(f'invalid Reed-Solomon basis {rs_basis}')
        if scrambler not in ['CCSDS', 'none']:
            raise ValueError(f'invalid scrambler {scrambler}')

        self.message_port_register_hier_out('out')

        if syncword_threshold is None:
            syncword_threshold = self.options.syncword_threshold

        self.slicer = digital.binary_slicer_fb()
        if precoding == 'differential':
            self.differential = digital.diff_decoder_bb(2)
        deframe_func = (sync_to_pdu if scrambler == 'CCSDS'
                        else sync_to_pdu_packed)
        packlen_mult = 8 if scrambler == 'CCSDS' else 1
        self.deframer = deframe_func(
            packlen=((frame_size + (32 * rs_interleaving if rs_en else 0)) *
                     packlen_mult),
            sync=_syncword, threshold=syncword_threshold)
        if scrambler == 'CCSDS':
            self.scrambler = ccsds_descrambler()
        self.fec = decode_rs(rs_basis == 'dual', rs_interleaving)

        self._blocks = [self, self.slicer]
        if precoding == 'differential':
            self._blocks.append(self.differential)
        self._blocks += [self.deframer]

        self.connect(*self._blocks)

        if rs_en:
            out = (self.fec, 'in')
        else:
            out = (self, 'out')

        if scrambler == 'CCSDS':
            self.msg_connect((self.deframer, 'out'), (self.scrambler, 'in'))
            self.msg_connect((self.scrambler, 'out'), out)
        else:
            self.msg_connect((self.deframer, 'out'), out)

        if rs_en:
            self.msg_connect((self.fec, 'out'), (self, 'out'))

    _default_sync_threshold = 4

    @classmethod
    def add_options(cls, parser):
        """
        Adds CCSDS RS deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--syncword_threshold', type=int,
            default=cls._default_sync_threshold,
            help='Syncword bit errors [default=%(default)r]')
        parser.add_argument(
            '--verbose_rs', action='store_true', help='Verbose RS decoder')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *

from ..adapters import *
from .csp import CSPHeader


Timestamp = UNIXTimestampAdapter(Int32sb)

Beacon = Struct(
    'csp_header' / CSPHeader,
    'beacon_counter' / Int16ub,
    'solar_panel_voltage' / LinearAdapter(1/16.0, Int8ub)[3],
    'eps_temp' / AffineAdapter(1, 100, Int8ub)[4],
    'eps_boot_cause' / Int8ub,
    'eps_batt_mode' / Int8ub,
    'solar_panel_current' / LinearAdapter(1/10.0, Int8ub),
    'system_input_current' / LinearAdapter(1/16.0, Int8ub),
    'battery_voltage' / LinearAdapter(1/34.0, Int8ub),
    'radio_PA_temp' / AffineAdapter(1, 100, Int8ub),
    'tx_count' / Int16ub,
    'rx_count' / Int16ub,
    'obc_temp' / AffineAdapter(1, 100, Int8ub)[2],
    'ang_velocity_mag' / Int8ub,
    'magnetometer' / LinearAdapter(1/6.0, Int8sb)[3],
    'main_axis_of_rot' / Int8ub
    )

WODBeacon = Struct(
    'csp_header' / CSPHeader,
    'timestamp' / Timestamp,
    'sp_voltage' / LinearAdapter(1000, Int16ub)[3],
    'sp_regulator_temp' / Int16sb[3],
    'battery_temp' / Int16sb,
    'boot_cause' / Int16ub,
    'battery_mode' / Int16ub,
    'solar_pannel_current' / Int16sb,
    'system_current' / Int16ub,
    'battery_voltage' / LinearAdapter(1000, Int16ub),
    'eps_boot_count' / Int16ub,
    'radio_amplifier_temp' / LinearAdapter(10, Int16sb),
    'tx_count' / Int16ub,
    'rx_count' / Int16ub,
    'last_rx_rf_power' / Int16sb,
    'last_rf_error' / Int16ub,
    'radio_boot_count' / Int16ub,
    'obc_temp' / LinearAdapter(10, Int16sb)[2],
    'gyro' / LinearAdapter(100, Int16sb)[3],
    'mag' / LinearAdapter(10, Int16sb)[3],
    'sp_regulator_current' / Int16ub[3],
    'sp_temp' / LinearAdapter(10, Int16sb)[6],
    'sun_sensor' / Int16sb[6]
    )


class sat_1kuns_pf:
    @staticmethod
    def parse(packet):
        if len(packet) == 138:
            # This is most likely an image packet
            return

        if len(packet) >= 92:
            return WODBeacon.parse(packet)
        else:
            return Beacon.parse(packet)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, digital

from ... import sx12xx_check_crc, cc11xx_packet_crop
from ...hier.sync_to_pdu_packed import sync_to_pdu_packed
from ...utils.options_block import options_block


# As syncword we use the last 0xAA from the preamble
# followed by the 0x7E synchronization marker
_syncword = '1010101001111110'


class binar1_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe the BINAR-1 custom framing

    This framing uses a custom syncword, a 1-byte packet length
    field, no scrambling, and CRC-16 CCITT FALSE (as in CCSDS
    transfer frames). the packet length field was supposed to
    always have the value 0x56, but we have seen other values
    such as 0x4b being used.

    The input is a float stream of soft symbols. The output are PDUs
    with frames.

    Args:
        syncword_threshold: number of bit errors allowed in syncword (int)
        options: Options from argparse
    """
    def __init__(self, syncword_threshold=None, options=None):
        gr.hier_block2.__init__(
            self,
            'binar1_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        if syncword_threshold is None:
            syncword_threshold = self.options.syncword_threshold

        self.slicer = digital.binary_slicer_fb()
        self.deframer = sync_to_pdu_packed(
            packlen=255+3, sync=_syncword, threshold=syncword_threshold)
        self.crop = cc11xx_packet_crop(use_crc16=True)
        self.crc = sx12xx_check_crc(self.options.verbose_crc, 0xffff, 0x0)

        self.connect(self, self.slicer, self.deframer)
        self.msg_connect((self.deframer, 'out'), (self.crop, 'in'))
        self.msg_connect((self.crop, 'out'), (self.crc, 'in'))
        self.msg_connect((self.crc, 'ok'), (self, 'out'))

    _default_sync_threshold = 0

    @classmethod
    def add_options(cls, parser):
        """
        Adds BINAR-1 deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--syncword_threshold', type=int,
            default=cls._default_sync_threshold,
            help='Syncword bit errors [default=%(default)r]')
        parser.add_argument(
            '--verbose_crc', action='store_true',
            help='Verbose CRC decoder')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *

from ..adapters import AffineAdapter, LinearAdapter

TMPrimaryHeader = BitStruct(
    'transfer_frame_version_number' / BitsInteger(2),
    'spacecraft_id' / BitsInteger(10),
    'virtual_channel_id' / BitsInteger(3),
    'ocf_flag' / Flag,
    'master_channel_frame_count' / BitsInteger(8),
    'virtual_channel_frame_count' / BitsInteger(8),
    'first_header_pointer' / BitsInteger(8)
    )

syncA = bytes([0x55, 0x55, 0x55, 0x55, 0x55, 0x55,
               0x00, 0x00, 0x08, 0x77, 0x80, 0x00, 0x00, 0x63])

syncB = bytes([0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x00, 0x26])

IVbat = LinearAdapter(2.5, Int16sb)
UVbat = LinearAdapter(2000.0, Int16sb)

Id = BitStruct(
    'other' / BitsInteger(13),
    'transponder' / Flag,
    'beacon' / Flag,
    'telemetry' / Flag
    )

hk_STM32_first_half = Struct(
    'id' / Id,
    'config' / Int8ub,
    'last_command' / Int8ub,
    'payload_mode' / Int8ub,
    'tx_mode' / Int8ub,
    'gain_tx' / Int16sb,
    'i_3v3' / LinearAdapter(10.0, Int16sb),
    'u_3v3' / LinearAdapter(2000.0, Int16sb),
    'i_vbat_tx' / IVbat,
    'u_vbat_tx' / UVbat,
    'i_vbat_rx' / IVbat,
    'u_vbat_rx' / UVbat,
    't_stm32' / AffineAdapter(1.0/0.29296875,
                              (304.0 - 25.0) / 0.29296875, Int16sb),
    't_pa' / LinearAdapter(8.0*16.0, Int16sb),
    'n_tx_rf' / Int16ub,
    'n_rx_rf' / Int16ub,
    'n_tx_err_rf' / Int16ub,
    'n_rx_err_rf' / Int16ub,
    'n_tx_can' / Int16ub,
    'n_rx_can' / Int16ub,
    'n_tx_err_can' / Int16ub,
    'n_rx_err_can' / Int16ub,
    'n_tc' / Int32ub,
    'dc_fm_tc' / Int16sb,
    'dc_fm_ham' / Int16sb,
    'rssi_fm_tc' / Int32ub,
    'rssi_fm_ham' / Int32ub,
    'reset_flag' / Int8ub,
    'sys_flag' / Int8ub,
    'dma_overflow' / Int16ub,
    'runtime_msb' / Int16ub
)

hk_STM32_second_half = Struct(
    'runtime_lsb' / Int16ub,
    'reset_count' / Int32ub,
    'ctcss_count' / Int32ub,
    'ctcss_det' / Float32b
    )

hk_AVR = Struct(
    'adf7021_ld' / Int8ub,
    'err_flag' / Int8ub,
    'callsign' / Bytes(6),
    'n_tx_232' / Int16ub,
    'n_rx_232' / Int16ub,
    'runtime' / Int32ub,
    'rssi_analog' / Int8ub,
    'n_rssi_const' / Int8ub,
    'unlock_count' / Int8ub,
    'reset_flag' / Int8ub,
    'reset_count' / Int32ub
    )

frameA = Struct(
    'header' / TMPrimaryHeader,
    Const(syncA),
    'hk_STM32_first' / hk_STM32_first_half
    )

frameB = Struct(
    'header' / TMPrimaryHeader,
    Const(syncB),
    'hk_STM32_second' / hk_STM32_second_half,
    'hk_AVR' / hk_AVR,
    'padding' / Bytes(30)
    )

framePadding = Struct(
    'header' / TMPrimaryHeader,
    'padding' / Bytes(76)
    )

by02 = Select(frameA, frameB, framePadding)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# BCH(15,k,d) implementation following https://en.wikipedia.org/wiki/BCH_code

import numpy as np

# Arithmetic in GF(16)

# exp_table[k] = a^k
exp_table = [8, 4, 2, 1, 12, 6, 3, 13, 10, 5, 14, 7, 15, 11, 9]
# j+1 = a^log_table[j]
log_table = [3, 2, 6, 1, 9, 5, 11, 0, 14, 8, 13, 4, 7, 10, 12]


def gf_mult(x, y):
    if x == 0 or y == 0:
        return 0
    return exp_table[(log_table[x-1] + log_table[y-1]) % len(exp_table)]


def gf_inv(x):
    if x == 0:
        raise(ValueError)
    return exp_table[-log_table[x-1] % len(exp_table)]


def compute_syndrome(p, j):
    # Syndrome calculation by polynomial evaluation
    s = 0
    n = 15
    for k in range(n - 1, -1, -1):
        if p & 1:
            s ^= exp_table[(k * j) % len(exp_table)]
        p >>= 1
    return s


def compute_error_locations(s):
    # Peterson-Gorenstein-Zierler algorithm
    L = [1, 0, 0, 0]  # coefficients of error locator polynomial
    if len(s) == 2:
        # This will raise ValueError if s[0] == 0, indicating decoding failure
        L[1] = gf_mult(s[1], gf_inv(s[0]))
    elif len(s) == 4:
        det_S = gf_mult(s[0], s[2]) ^ gf_mult(s[1], s[1])
        if det_S == 0:
            # Matrix is non-invertible. Throw away 2 syndromes.
            return compute_error_locations(s[:-2])
        inv_det_S = gf_inv(det_S)
        L[2] = gf_mult(s[2], s[2]) ^ gf_mult(s[3], s[1])
        L[2] = gf_mult(L[2], inv_det_S)
        L[1] = gf_mult(s[0], s[3]) ^ gf_mult(s[2], s[1])
        L[1] = gf_mult(L[1], inv_det_S)
    elif len(s) == 6:
        det_S = (
            gf_mult(gf_mult(s[0], s[2]), s[4])
            ^ gf_mult(gf_mult(s[2], s[2]), s[2])
            ^ gf_mult(gf_mult(s[1], s[1]), s[4])
            ^ gf_mult(gf_mult(s[3], s[3]), s[0]))
        if det_S == 0:
            # Matrix is non-invertible. Throw away 2 syndromes.
            return compute_error_locations(s[:-2])
        inv_det_S = gf_inv(det_S)
        L[3] = (
            gf_mult(gf_mult(s[3], s[2]), s[4])
            ^ gf_mult(gf_mult(s[1], s[3]), s[5])
            ^ gf_mult(gf_mult(s[4], s[3]), s[2])
            ^ gf_mult(gf_mult(s[2], s[2]), s[5])
            ^ gf_mult(gf_mult(s[1], s[4]), s[4])
            ^ gf_mult(gf_mult(s[3], s[3]), s[3]))
        L[3] = gf_mult(L[3], inv_det_S)
        L[2] = (
            gf_mult(gf_mult(s[0], s[4]), s[4])
            ^ gf_mult(gf_mult(s[1], s[2]), s[5])
            ^ gf_mult(gf_mult(s[3], s[3]), s[2])
            ^ gf_mult(gf_mult(s[2], s[2]), s[4])
            ^ gf_mult(gf_mult(s[1], s[3]), s[4])
            ^ gf_mult(gf_mult(s[0], s[3]), s[5]))
        L[2] = gf_mult(L[2], inv_det_S)
        L[1] = (
            gf_mult(gf_mult(s[0], s[2]), s[5])
            ^ gf_mult(gf_mult(s[1], s[3]), s[3])
            ^ gf_mult(gf_mult(s[4], s[1]), s[2])
            ^ gf_mult(gf_mult(s[2], s[2]), s[3])
            ^ gf_mult(gf_mult(s[1], s[1]), s[5])
            ^ gf_mult(gf_mult(s[0], s[3]), s[4]))
        L[1] = gf_mult(L[1], inv_det_S)
    # Brute-force search roots of error locator polynomial
    return [j for j in range(15) if
            exp_table[0]
            ^ gf_mult(L[1], exp_table[-j % len(exp_table)])
            ^ gf_mult(L[2], exp_table[-2*j % len(exp_table)])
            ^ gf_mult(L[3], exp_table[-3*j % len(exp_table)])
            == 0]


def decode_bch15(bits, d=7):
    """BCH(15,k,d) decode function

The following values of (n,k,d) are supported:
(15,11,3), (15,7,5), (15,5,7)

Expects an np.array() as bits and modifies it in place
returns True if decode is successful
"""
    b = np.packbits(bits)
    p = (b[0] << 7) | (b[1] >> 1)
    syndromes = [compute_syndrome(p, j) for j in range(d - 1)]
    if not any(syndromes):
        # If all syndromes are zero, there are no errors
        return True
    try:
        errors = compute_error_locations(syndromes)
    except ValueError:
        return False
    for e in errors:
        bits[e] ^= 1
    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import struct


class CSP(object):
    def __init__(self, csp_packet):
        if len(csp_packet) < 4:
            raise ValueError('Malformed CSP packet (too short)')
        csp = struct.unpack('<I', csp_packet[0:4])[0]
        self.priority = (csp >> 30) & 0x3
        self.source = (csp >> 25) & 0x1f
        self.destination = (csp >> 20) & 0x1f
        self.dest_port = (csp >> 14) & 0x3f
        self.source_port = (csp >> 8) & 0x3f
        self.reserved = (csp >> 4) & 0xf
        self.hmac = (csp >> 3) & 1
        self.xtea = (csp >> 2) & 1
        self.rdp = (csp >> 1) & 1
        self.crc = csp & 1

    def __str__(self):
        return ("""CSP header:
        Priority:\t\t{}
        Source:\t\t\t{}
        Destination:\t\t{}
        Destination port:\t{}
        Source port:\t\t{}
        Reserved field:\t\t{}
        HMAC:\t\t\t{}
        XTEA:\t\t\t{}
        RDP:\t\t\t{}
        CRC:\t\t\t{}""".format(
            self.priority, self.source, self.destination, self.dest_port,
            self.source_port, self.reserved, self.hmac, self.xtea, self.rdp,
            self.crc))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr
import pmt

from ... import filereceiver


class file_receiver(gr.basic_block):
    """
    Block for file reception

    The input are PDUs with frames

    The frames are expected to contain file chunks, which are saved
    into files using a FileReceiver object.

    Args:
        receiver: FileReceiver object (to load from the
            filereceiver package) (str)
        path: path to save files to (str)
        verbose: use verbose messages in FileReceiver (bool)
        options: options from argparse
        **kwargs: these are passed straight to the FileReceiver object
    """
    def __init__(self, receiver, path=None, verbose=None,
                 options=None, **kwargs):
        gr.basic_block.__init__(
            self,
            'file_receiver',
            in_sig=[],
            out_sig=[])
        if verbose is None:
            if options is not None:
                verbose = options.verbose_file_receiver
            else:
                raise ValueError(
                    'Must indicate verbose in function arguments or options')
        if path is None:
            if options is not None:
                path = options.file_output_path
            else:
                raise ValueError(
                    'Must indicate path in function arguments or options')
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.receiver = getattr(filereceiver, receiver)(path,
                                                        verbose, **kwargs)

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = bytes(pmt.u8vector_elements(msg))

        self.receiver.push_chunk(packet)

    @classmethod
    def add_options(cls, parser):
        """
        Adds telemetry parser specific options to the argparse parser
        """
        parser.add_argument(
            '--file_output_path', default='.',
            help='File output path [default=%(default)r]')
        parser.add_argument(
            '--verbose_file_receiver',
            action='store_true',
            help='Verbose file receiver')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks

from .ccsds_rs_deframer import ccsds_rs_deframer
from ...hier.ccsds_viterbi import ccsds_viterbi
from ...utils.options_block import options_block


class ccsds_concatenated_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe CCSDS concatenated
    (convolutional + Reed-Solomon) TM frames

    The input is a float stream of soft symbols. The output are PDUs
    with frames.

    Args:
        frame_size: frame size (not including parity check bytes) (int)
        precoding: either None or 'differential' for differential precoding
                   (str)
        rs_en: If Reed-Solomon should be enabled or not (bool)
        rs_basis: Reed-Solomon basis, either 'conventional' or 'dual' (str)
        rs_interleaving: Reed-Solomon interleaving depth (int)
        scrambler: scrambler to use, either 'CCSDS' or 'none' (str)
        convolutional: convolutional code to use, one of the following:
            'CCSDS', 'NASA-DSN', 'CCSDS uninverted',
            'NASA-DSN uninverted' (str)
        syncword_threshold: number of bit errors allowed in syncword (int)
        options: Options from argparse
    """
    def __init__(self, frame_size=223, precoding=None, rs_en=True,
                 rs_basis='dual', rs_interleaving=1, scrambler='CCSDS',
                 convolutional='CCSDS', syncword_threshold=None, options=None):
        gr.hier_block2.__init__(
            self,
            'ccsds_concatenated_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        self.delay1 = blocks.delay(gr.sizeof_float, 1)
        self.viterbi0 = ccsds_viterbi(convolutional)
        self.viterbi1 = ccsds_viterbi(convolutional)
        self.char2float0 = blocks.char_to_float(1, 1)
        self.char2float1 = blocks.char_to_float(1, 1)
        self.addconst0 = blocks.add_const_ff(-0.5)
        self.addconst1 = blocks.add_const_ff(-0.5)
        self.rs0 = ccsds_rs_deframer(
            frame_size, precoding, rs_en, rs_basis, rs_interleaving,
            scrambler, syncword_threshold, options)
        self.rs1 = ccsds_rs_deframer(
            frame_size, precoding, rs_en, rs_basis, rs_interleaving,
            scrambler, syncword_threshold, options)

        self.connect(self, self.viterbi0, self.char2float0,
                     self.addconst0, self.rs0)
        self.connect(self, self.delay1, self.viterbi1, self.char2float1,
                     self.addconst1, self.rs1)
        self.msg_connect((self.rs0, 'out'), (self, 'out'))
        self.msg_connect((self.rs1, 'out'), (self, 'out'))

    @classmethod
    def add_options(cls, parser):
        """
        Adds CCSDS concatenated deframer options to the argparse parser
        """
        ccsds_rs_deframer.add_options(parser)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import struct

from gnuradio import gr
import numpy as np
import pmt


# This block is based on the Universal SPUNTIX Protocol (USP)
# Protocol Description revision 1.04
# https://sputnix.ru/tpl/docs/amateurs/USP%20protocol%20description%20v1.04.pdf

class usp_ax25_crop(gr.basic_block):
    """
    Crop a USP packet to obtain the encapsulated AX.25 frame
    """
    def __init__(self):
        gr.basic_block.__init__(
            self,
            name='usp_ax25_crop',
            in_sig=[],
            out_sig=[])
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        msg = pmt.u8vector_elements(msg)

        length_field = msg[2:4]
        length = struct.unpack('<H', bytes(length_field))[0]

        msg_out = msg[4:][:length]

        msg_out = pmt.init_u8vector(len(msg_out), msg_out)
        msg_out = pmt.cons(pmt.car(msg_pmt), msg_out)
        self.message_port_pub(pmt.intern('out'), msg_out)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *


SSID = BitStruct(
    'ch' / Flag,  # C / H bit
    Default(BitsInteger(2), 3),  # reserved bits
    'ssid' / BitsInteger(4),
    'extension' / Flag  # last address bit
    )


class CallsignAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return bytes([x << 1 for x in bytes(
            (obj.upper() + ' '*6)[:6], encoding='ascii')])

    def _decode(self, obj, context, path=None):
        return str(bytes([x >> 1 for x in obj]), encoding='ascii').strip()


Callsign = CallsignAdapter(Bytes(6))

Address = Struct(
        'callsign' / Callsign,
        'ssid' / SSID
        )

Control = Hex(Int8ub)
Control16 = Hex(Int16ub)

PID = Hex(Int8ub)

Header = Struct(
    'addresses' / RepeatUntil(lambda x, lst, ctx: x.ssid.extension, Address),
    'control' / Control,
    'pid' / PID
    )

Header16 = Struct(
    'addresses' / RepeatUntil(lambda x, lst, ctx: x.ssid.extension, Address),
    'control' / Control16,
    'pid' / PID
    )

Frame = Struct(
    'header' / Header,
    'info' / GreedyBytes
    )

ax25 = Frame



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *


TMSecondaryHeader = BitStruct(
    'version' / BitsInteger(4),
    'time_reference_status' / BitsInteger(4),
    'service_type_id' / BitsInteger(8),
    'message_subtype_id' / BitsInteger(8),
    'message_type_counter' / BitsInteger(16),
    'destination_id' / BitsInteger(16))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import os
import sys

from construct.core import ConstructError
from gnuradio import gr
import pmt

from ... import telemetry


class telemetry_parser(gr.basic_block):
    """
    Block for telemetry parsing

    The input are PDUs with frames

    These are parsed with construct and the text output is printed
    or saved to file

    Args:
        definition: telemetry definition name (to load from the telemetry
            package) (str)
        file: file or file path to print output (defaults to sys.stdout)
        options: options from argparse
    """
    def __init__(self, definition, file=sys.stdout, options=None):
        gr.basic_block.__init__(
            self,
            'telemetry_parser',
            in_sig=[],
            out_sig=[])
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.format = getattr(telemetry, definition)
        if getattr(options, 'telemetry_output', None):
            file = options.telemetry_output
        if type(file) in [str, bytes] or isinstance(file, os.PathLike):
            file = open(file, 'a')
        self.file = file

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = bytes(pmt.u8vector_elements(msg))

        meta = pmt.car(msg_pmt)
        transmitter = pmt.dict_ref(
            meta, pmt.intern('transmitter'), pmt.PMT_NIL)
        if pmt.is_symbol(transmitter):
            print('-> Packet from', pmt.symbol_to_string(transmitter),
                  file=self.file)

        try:
            data = self.format.parse(packet)
        except ConstructError as e:
            print(f'Could not parse telemetry beacon {e}', file=self.file)
            return
        if data:
            print(data, file=self.file)

    @classmethod
    def add_options(cls, parser):
        """
        Adds telemetry parser specific options to the argparse parser
        """
        parser.add_argument(
            '--telemetry_output', default=sys.stdout,
            help='Telemetry output file [default=stdout]')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr
import numpy
import pmt


crc_table = [
    0x0000, 0xc0c1, 0xc181, 0x0140, 0xc301, 0x03c0, 0x0280, 0xc241,
    0xc601, 0x06c0, 0x0780, 0xc741, 0x0500, 0xc5c1, 0xc481, 0x0440,
    0xcc01, 0x0cc0, 0x0d80, 0xcd41, 0x0f00, 0xcfc1, 0xce81, 0x0e40,
    0x0a00, 0xcac1, 0xcb81, 0x0b40, 0xc901, 0x09c0, 0x0880, 0xc841,
    0xd801, 0x18c0, 0x1980, 0xd941, 0x1b00, 0xdbc1, 0xda81, 0x1a40,
    0x1e00, 0xdec1, 0xdf81, 0x1f40, 0xdd01, 0x1dc0, 0x1c80, 0xdc41,
    0x1400, 0xd4c1, 0xd581, 0x1540, 0xd701, 0x17c0, 0x1680, 0xd641,
    0xd201, 0x12c0, 0x1380, 0xd341, 0x1100, 0xd1c1, 0xd081, 0x1040,
    0xf001, 0x30c0, 0x3180, 0xf141, 0x3300, 0xf3c1, 0xf281, 0x3240,
    0x3600, 0xf6c1, 0xf781, 0x3740, 0xf501, 0x35c0, 0x3480, 0xf441,
    0x3c00, 0xfcc1, 0xfd81, 0x3d40, 0xff01, 0x3fc0, 0x3e80, 0xfe41,
    0xfa01, 0x3ac0, 0x3b80, 0xfb41, 0x3900, 0xf9c1, 0xf881, 0x3840,
    0x2800, 0xe8c1, 0xe981, 0x2940, 0xeb01, 0x2bc0, 0x2a80, 0xea41,
    0xee01, 0x2ec0, 0x2f80, 0xef41, 0x2d00, 0xedc1, 0xec81, 0x2c40,
    0xe401, 0x24c0, 0x2580, 0xe541, 0x2700, 0xe7c1, 0xe681, 0x2640,
    0x2200, 0xe2c1, 0xe381, 0x2340, 0xe101, 0x21c0, 0x2080, 0xe041,
    0xa001, 0x60c0, 0x6180, 0xa141, 0x6300, 0xa3c1, 0xa281, 0x6240,
    0x6600, 0xa6c1, 0xa781, 0x6740, 0xa501, 0x65c0, 0x6480, 0xa441,
    0x6c00, 0xacc1, 0xad81, 0x6d40, 0xaf01, 0x6fc0, 0x6e80, 0xae41,
    0xaa01, 0x6ac0, 0x6b80, 0xab41, 0x6900, 0xa9c1, 0xa881, 0x6840,
    0x7800, 0xb8c1, 0xb981, 0x7940, 0xbb01, 0x7bc0, 0x7a80, 0xba41,
    0xbe01, 0x7ec0, 0x7f80, 0xbf41, 0x7d00, 0xbdc1, 0xbc81, 0x7c40,
    0xb401, 0x74c0, 0x7580, 0xb541, 0x7700, 0xb7c1, 0xb681, 0x7640,
    0x7200, 0xb2c1, 0xb381, 0x7340, 0xb101, 0x71c0, 0x7080, 0xb041,
    0x5000, 0x90c1, 0x9181, 0x5140, 0x9301, 0x53c0, 0x5280, 0x9241,
    0x9601, 0x56c0, 0x5780, 0x9741, 0x5500, 0x95c1, 0x9481, 0x5440,
    0x9c01, 0x5cc0, 0x5d80, 0x9d41, 0x5f00, 0x9fc1, 0x9e81, 0x5e40,
    0x5a00, 0x9ac1, 0x9b81, 0x5b40, 0x9901, 0x59c0, 0x5880, 0x9841,
    0x8801, 0x48c0, 0x4980, 0x8941, 0x4b00, 0x8bc1, 0x8a81, 0x4a40,
    0x4e00, 0x8ec1, 0x8f81, 0x4f40, 0x8d01, 0x4dc0, 0x4c80, 0x8c41,
    0x4400, 0x84c1, 0x8581, 0x4540, 0x8701, 0x47c0, 0x4680, 0x8641,
    0x8201, 0x42c0, 0x4380, 0x8341, 0x4100, 0x81c1, 0x8081, 0x4040]


def crc16_arc(data):
    crc = 0

    for byte in data:
        tbl_idx = (crc ^ byte) & 0xff
        crc = (crc_table[tbl_idx] ^ (crc >> 8)) & 0xffff

    return crc & 0xffff


class check_tt64_crc(gr.basic_block):
    """docstring for block check_tt64_crc"""
    def __init__(self, verbose, packet_len=48, strip=True):
        gr.basic_block.__init__(
            self,
            name='check_tt64_crc',
            in_sig=[],
            out_sig=[])

        self.verbose = verbose
        self.packet_len = packet_len
        self.strip = strip

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('ok'))
        self.message_port_register_out(pmt.intern('fail'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = bytes(pmt.u8vector_elements(msg))

        if self.packet_len is not None:
            packet = packet[:self.packet_len]

        if ((self.packet_len is not None and len(packet) < self.packet_len)
                or (self.packet_len is None and len(packet) < 2)):
            if self.verbose:
                print('Packet too short')
            return

        packet_out = packet[:-2] if self.strip else packet
        msg_out = pmt.cons(pmt.PMT_NIL,
                           pmt.init_u8vector(len(packet_out), packet_out))
        if crc16_arc(packet) == 0:
            if self.verbose:
                print('CRC OK')
            self.message_port_pub(pmt.intern('ok'), msg_out)
        else:
            if self.verbose:
                print('CRC failed')
            self.message_port_pub(pmt.intern('fail'), msg_out)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018,2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr
import numpy as np
import pmt

from .bch15 import decode_bch15
from .telemetry.snet import LTUFrameHeader


class snet_deframer(gr.basic_block):
    """docstring for block snet_deframer"""
    def __init__(self, verbose=False, buggy_crc=True):
        gr.basic_block.__init__(
            self,
            name='snet_deframer',
            in_sig=[],
            out_sig=[])
        self.verbose = verbose
        self.buggy_crc = buggy_crc

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        bits = np.array(pmt.u8vector_elements(msg))

        ltu = bits[:210].reshape((15, 14)).transpose()

        # Decode BCH(15,5,7)
        if not all((decode_bch15(ltu[j, :]) for j in range(14))):
            # Decode failure
            if self.verbose:
                print('BCH decode failure')
            return

        ltu = np.fliplr(ltu[:, -5:]).ravel()
        hdr = LTUFrameHeader.parse(np.packbits(ltu))

        ltu_crc = np.concatenate((ltu[:-5], np.array([1, 0, 1, 1, 0, 1, 1])))
        ltu_crc = ltu_crc.reshape((9, 8))
        if self.buggy_crc:
            # Reverse byte ordering for CRC5 calculation
            ltu_crc = np.flipud(ltu_crc)

        if self.buggy_crc:
            # Force CRC5 bugs
            ltu_crc[4, :] = ltu_crc[3, :]
        # CRC5 calculation
        crc = 0x1F
        for bit in ltu_crc.ravel():
            # Check most significant bit in the CRC buffer and save
            # in a variable.
            c = crc & 0x10
            # Shift variable to make the compare op. possible (see beneath).
            c >>= 4
            # Shift CRC to the left and write 0 into the least significant bit.
            crc <<= 1
            if c != bit:
                crc ^= 0x15  # CRC polynomial
            crc &= 0x1F

        if crc != hdr.CRC5:
            if self.verbose:
                print('CRC5 fail')
            return

        if self.verbose:
            print(hdr)

        if hdr.PduLength == 0:
            return

        codewords_per_block = 16
        uncoded = False
        if hdr.AiTypeSrc == 0:
            uncoded = True
        elif hdr.AiTypeSrc == 1:
            data_bits_per_codeword = 11  # BCH(15,11,3)
            bch_d = 3
        elif hdr.AiTypeSrc == 2:
            data_bits_per_codeword = 7  # BCH(15,7,5)
            bch_d = 5
        elif hdr.AiTypeSrc == 3:
            data_bits_per_codeword = 5  # BCH(15,5,7)
            bch_d = 7
        else:
            if self.verbose:
                print('Invalid AiTypeSrc')
            return

        if uncoded:
            pdu_bytes = bits[210:210+hdr.PduLength*8]
            pdu_bytes = pdu_bytes.reshape((hdr.PduLength, 8))
            pdu_bytes = np.fliplr(pdu_bytes)
        else:
            data_bytes_per_block = (codewords_per_block
                                    * data_bits_per_codeword // 8)
            num_blocks = int(np.ceil(float(
                hdr.PduLength) / data_bytes_per_block))

            blocks = list()
            for k in range(num_blocks):
                block = bits[210+k*16*15:210+(k+1)*16*15].reshape((15, 16))
                if bch_d:
                    block = block.transpose()

                if not bch_d:
                    print(block)

                # Decode BCH
                if (bch_d
                        and not all((decode_bch15(block[j, :], d=bch_d)
                                     for j in range(16)))):
                    # Decode failure
                    if self.verbose:
                        print('BCH decode failure')
                    return

                if bch_d:
                    blocks.append(block[:, -data_bits_per_codeword:].ravel())
                else:
                    blocks.append(block.ravel())

            pdu_bytes = np.concatenate(blocks)
            pdu_bytes = pdu_bytes.reshape((data_bytes_per_block
                                           * num_blocks, 8))
            pdu_bytes = np.fliplr(pdu_bytes)
            # Drop 0xDB padding bytes at the end
            pdu_bytes = pdu_bytes[:hdr.PduLength]
            if not bch_d:
                print(pdu_bytes)

        # CRC13
        crc = 0x1FFF
        pdu_crc = np.flipud(pdu_bytes) if self.buggy_crc else pdu_bytes
        for bit in pdu_crc.ravel():
            # Check most significant bit in the CRC buffer and save it
            # in a variable.
            c = crc & 0x1000
            # Shift variable to make the compare op. possible (see beneath).
            c >>= 12
            # Shift CRC to the left and write 0 into the least significant bit.
            crc <<= 1
            if (c or bit if self.buggy_crc else c != bit):
                crc ^= 0x1CF5  # CRC polynomial
            crc &= 0x1FFF

        if crc != hdr.CRC13:
            if self.verbose:
                print('CRC13 fail')
            return

        pdu = bytes(np.packbits(pdu_bytes))
        pdu_tags = pmt.make_dict()
        pdu_tags = pmt.dict_add(
            pdu_tags, pmt.intern('SNET SrcId'), pmt.from_long(hdr.SrcId))
        self.message_port_pub(
            pmt.intern('out'),
            pmt.cons(pdu_tags, pmt.init_u8vector(len(pdu), pdu)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import numpy


FEND = numpy.uint8(0xc0)
FESC = numpy.uint8(0xdb)
TFEND = numpy.uint8(0xdc)
TFESC = numpy.uint8(0xdd)


def kiss_escape(a):
    """Escapes KISS control characters

    This replaces FEND and FESC according to the KISS escape rules
    """
    buff = list()
    for x in a:
        if x == FESC:
            buff.append(FESC)
            buff.append(TFESC)
        elif x == FEND:
            buff.append(FESC)
            buff.append(TFEND)
        else:
            buff.append(numpy.uint8(x))
    return buff



# -*- coding: utf-8 -*-

#
# SPDX-License-Identifier: GPL-3.0
#
# GNU Radio Python Flow Graph
# Title: SI4463 scrambler
# Author: Daniel Estevez
# Description: SI4463 scrambler
# GNU Radio version: 3.8.0.0

from gnuradio import blocks
from gnuradio import digital
from gnuradio import gr
from gnuradio.filter import firdes


class si4463_scrambler(gr.hier_block2):
    def __init__(self):
        gr.hier_block2.__init__(
            self,
            'SI4463 scrambler',
            gr.io_signature(0, 0, 0),
            gr.io_signature(0, 0, 0),
        )
        self.message_port_register_hier_in('in')
        self.message_port_register_hier_out('out')

        ##################################################
        # Blocks
        ##################################################
        self.digital_additive_scrambler_bb_0_0 = (
            digital.additive_scrambler_bb(
                0x21, 0x1e1, 8, count=0, bits_per_byte=1,
                reset_tag_key="packet_len"))
        self.blocks_tagged_stream_to_pdu_0_0 = (
            blocks.tagged_stream_to_pdu(blocks.byte_t, 'packet_len'))
        self.blocks_tagged_stream_multiply_length_0 = (
            blocks.tagged_stream_multiply_length(gr.sizeof_char*1,
                                                 'packet_len', 1.0/8))
        self.blocks_pdu_to_tagged_stream_0 = (
            blocks.pdu_to_tagged_stream(blocks.byte_t, 'packet_len'))
        self.blocks_pack_k_bits_bb_0 = blocks.pack_k_bits_bb(8)

        ##################################################
        # Connections
        ##################################################
        self.msg_connect(
            (self.blocks_tagged_stream_to_pdu_0_0, 'pdus'), (self, 'out'))
        self.msg_connect(
            (self, 'in'), (self.blocks_pdu_to_tagged_stream_0, 'pdus'))
        self.connect(
            (self.blocks_pack_k_bits_bb_0, 0),
            (self.blocks_tagged_stream_multiply_length_0, 0))
        self.connect(
            (self.blocks_pdu_to_tagged_stream_0, 0),
            (self.digital_additive_scrambler_bb_0_0, 0))
        self.connect(
            (self.blocks_tagged_stream_multiply_length_0, 0),
            (self.blocks_tagged_stream_to_pdu_0_0, 0))
        self.connect(
            (self.digital_additive_scrambler_bb_0_0, 0),
            (self.blocks_pack_k_bits_bb_0, 0))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import numpy as np
from gnuradio import gr
import pmt


# Details of the NGHam protocol taken from
# https://github.com/skagmo/ngham/blob/master/ngham.c

_tags = [
    0b001110110100100111001101,
    0b010011011101101001010111,
    0b011101101001001110011010,
    0b100110111011010010101110,
    0b101000001111110101100011,
    0b110101100110111011111001,
    0b111011010010011100110100]
ngham_tags = np.array([list(map(int, ('0'*24 + bin(t)[2:])[-24:]))
                       for t in _tags],
                      dtype='uint8')
ngham_rs_sizes = [47, 79, 111, 159, 191, 223, 255]


class ngham_packet_crop(gr.basic_block):
    """
    Crops an NGHam packet according to its size tag

    Input: An NGHam packet with size tag and RS codeword
    Output: The RS codeword appropriately cropped, routed to
    rs16 or rs32 according to the RS parity check bytes
    """
    def __init__(self):
        gr.basic_block.__init__(
            self,
            name='ngham_packet_handler',
            in_sig=[],
            out_sig=[])
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('rs16'))
        self.message_port_register_out(pmt.intern('rs32'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = bytes(pmt.u8vector_elements(msg))

        size_tag = np.unpackbits(np.frombuffer(packet[:3], dtype='uint8'))
        tag_dist = np.sum(size_tag ^ ngham_tags, axis=1)
        best_tag = np.argmin(tag_dist)

        rs_size = ngham_rs_sizes[best_tag]
        if len(packet[3:]) < rs_size:
            print('Packet is too short')
            return

        out = 'rs16' if best_tag < 3 else 'rs32'
        self.message_port_pub(
            pmt.intern(out),
            pmt.cons(pmt.car(msg_pmt),
                     pmt.init_u8vector(rs_size, packet[3:3+rs_size])))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017,2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr
import numpy
import pmt


class sat_3cat_2:
    def parse(self, packet):
        data = packet[17:].split()

        status = {b'1': 'Survival', b'2': 'Sun-safe', b'3': 'Nominal',
                  b'4': 'TX', b'5': 'RX', b'6': 'Payload', b'7': 'Payload'}
        adcs = {b'0': 'auto', b'1': 'manual'}
        if data[5] == b'0':
            detumbling = (
                'Detumbling  ({},{},{})nT'.format(
                    float(data[7]), float(data[8]), float(data[9])))
        else:
            detumbling = (
                'SS-nominal  Sun: ({:.2f},{:.2f},{:.2f})'.format(
                    float(data[7]), float(data[8]), float(data[9])))

        string = (
            status[data[0]]
            + '  {:.2f}V  {}mA'.format(int(data[1])/1000.0, int(data[2])) +
            '  EPS: {}ºC   Ant: {}ºC'.format(data[3], data[4]) +
            '  ADCS ' + adcs[data[6]] + '  ' + detumbling +
            '  Control: ({:.1e},{:.1e},{:.1e})V'.format(
                float(data[10]), float(data[11]), float(data[12])))

        return string



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, digital

from ... import ax100_decode, u482c_decode
from ...hier.sync_to_pdu_packed import sync_to_pdu_packed
from ...utils.options_block import options_block


_syncword = '10010011000010110101000111011110'


class ax100_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe the GOMspace AX100 protocols.

    The input is a float stream of soft symbols. The output are PDUs
    with frames (most likely CSP frames).

    Args:
        mode: mode to use ('RS' or 'ASM') (string)
        scrambler: scrambler to use, either 'CCSDS' or 'none'
                   (only for ASM mode) (str)
        syncword_threshold: number of bit errors allowed in syncword (int)
        options: Options from argparse
    """
    def __init__(self, mode, scrambler='CCSDS', syncword_threshold=None,
                 options=None):
        gr.hier_block2.__init__(
            self,
            'ax100_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        if scrambler not in ['CCSDS', 'none']:
            raise ValueError(f'invalid scrambler {scrambler}')

        self.message_port_register_hier_out('out')

        if syncword_threshold is None:
            syncword_threshold = self.options.syncword_threshold

        if mode not in ['RS', 'ASM']:
            raise Exception("Unsupported AX100 mode. Use 'RS' or 'ASM'")

        self.slicer = digital.binary_slicer_fb()
        if mode == 'RS':
            self.descrambler = digital.descrambler_bb(0x21, 0, 16)
        self.deframer = sync_to_pdu_packed(
            packlen=256 if mode == 'RS' else 255,
            sync=_syncword, threshold=syncword_threshold)
        self.fec = (ax100_decode(self.options.verbose_fec) if mode == 'RS'
                    else u482c_decode(
                        self.options.verbose_fec, 0,
                        1 if scrambler == 'CCSDS' else 0, 1))
        self._blocks = [self, self.slicer]
        if mode == 'RS':
            self._blocks.append(self.descrambler)
        self._blocks += [self.deframer]

        self.connect(*self._blocks)
        self.msg_connect((self.deframer, 'out'), (self.fec, 'in'))
        self.msg_connect((self.fec, 'out'), (self, 'out'))

    _default_sync_threshold = 4

    @classmethod
    def add_options(cls, parser):
        """
        Adds AX100 deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--syncword_threshold', type=int,
            default=cls._default_sync_threshold,
            help='Syncword bit errors [default=%(default)r]')
        parser.add_argument(
            '--verbose_fec', action='store_true', help='Verbose FEC decoder')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import collections

import numpy
from gnuradio import gr
import pmt


class fixedlen_tagger(gr.basic_block):
    """
    docstring for block fixedlen_tagger
    """
    def __init__(self, syncword_tag, packetlen_tag, packet_len, stream_type):
        gr.basic_block.__init__(
            self,
            name='fixedlen_tagger',
            in_sig=[stream_type],
            out_sig=[stream_type])
        self.syncword_tag = pmt.string_to_symbol(syncword_tag)
        self.packetlen_tag = pmt.string_to_symbol(packetlen_tag)
        self.packet_len = packet_len
        self.stream = collections.deque(maxlen=packet_len - 1)
        self.maxtag = -1
        self.data = []
        self.tags = []
        self.tags_to_write = []
        self.written = 0
        self.really_written = 0

    def set_packet_len(self, packet_len):
        self.packet_len = packet_len
        self.stream = collections.deque(
            self.stream, maxlen=self.packet_len - 1)

    def try_to_flush(self, out):
        # Try to send as much items as we have in buffer
        len_write = min(len(self.data), len(out))
        out[:len_write] = self.data[:len_write]
        self.data = self.data[len_write:]
        self.really_written += len_write

        for tag in self.tags_to_write[:]:
            # Modifying self.tags_to_write during the loop would
            # disturb the loop control, so we copy it
            if tag < self.really_written:
                self.tags_to_write.remove(tag)
                self.add_item_tag(0, tag, self.packetlen_tag,
                                  pmt.from_long(self.packet_len))

        return len_write

    def general_work(self, input_items, output_items):
        inp = input_items[0]
        out = output_items[0]

        if self.data:
            # Write as much as possible without consuming input
            return self.try_to_flush(out)

        window = list(self.stream) + inp.tolist()

        alltags = self.get_tags_in_range(0, self.maxtag + 1,
                                         self.nitems_read(0) + len(inp),
                                         self.syncword_tag)
        for tag in alltags:
            if tag.offset not in self.tags:
                self.maxtag = max(self.maxtag, tag.offset)
                self.tags.append(tag.offset)
        for tag in self.tags[:]:
            # Modifying self.tags during the loop would disturb the
            # loop control, so we copy it
            if (tag >= self.nitems_read(0) - len(self.stream)
                    and
                    tag < self.nitems_read(0) + len(inp)
                    - self.packet_len + 1):
                self.tags.remove(tag)
                start = tag - self.nitems_read(0) + len(self.stream)
                packet = window[start:start+self.packet_len]
                self.data += packet
                self.tags_to_write.append(self.written)
                self.written += self.packet_len

        self.stream.extend(inp.tolist())

        self.consume(0, len(inp))
        return self.try_to_flush(out)



from ..adapters import LinearAdapter
from construct import Adapter, BitsInteger, BitStruct, Container, Enum, Flag, \
                      Int8ub, Int16ub, Int32ub, Padding, Struct, Switch

SOH_L0 = BitStruct(
    'SOH_L0__WDT_2SEC_CNT' / BitsInteger(3),
    'SOH_L0__RESET_ARMED' / Enum(Flag, ARMED=1, NOT_ARMED=0),
    'SOH_L0__WDT_STAT' / Enum(Flag, NO_WDT=0, WDT=1),
    'SOH_L0__WDT_EN' / Enum(Flag, DISABLED=0, ENABLED=1),
    'SOH_L0__TABLE_SELECT' / Enum(Flag, FLASH=0, COMPILED=1),
    'SOH_L0__BOOT_RELAY' / Enum(Flag, PRIMARY=1, REDUNDANT=0),
    'SOH_L0__L0_ACPT_CNT' / BitsInteger(8),
    'SOH_L0__L0_RJCT_CNT' / BitsInteger(8),
    'SOH_L0__HW_SEC_CNT' / BitsInteger(8),
    Padding(64),
    'SOH_L0__TIME_TAG' / BitsInteger(32),
    Padding(32),
    'SOH_L0__PLD_TLM_ACK_CNT' / BitsInteger(8),
    'SOH_L0__PLD_CMD_CNT' / BitsInteger(8),
    'SOH_L0__PLD_TLM_TO_CNT' / BitsInteger(8),
    'SOH_L0__PLD_TLM_NAK_CNT' / BitsInteger(8),
    'SOH_L0__SPARE_END' / BitsInteger(64),
)

SOH_COMMAND_TLM = BitStruct(
    'SOH_COMMAND_TLM__CMD_STATUS' / Enum(BitsInteger(8), OK=0, BAD_APID=1,
                                         BAD_OPCODE=2, BAD_DATA=3,
                                         NOW_READING=4, DONE_READING=5, IDLE=6,
                                         NO_CMD_DATA=7, CMD_SRVC_OVERRUN=8,
                                         CMD_APID_OVERRUN=9,
                                         INCORRECT_WHEEL_MODE=10,
                                         BAD_ELEMENT=11, TABLES_BUSY=12,
                                         FLASH_NOT_ARMED=13,
                                         THRUSTERS_DISABLED=14,
                                         ATT_ERR_TOO_HIGH=15, ASYNC_REFUSED=16,
                                         DRIVER_ERROR=17),
    'SOH_COMMAND_TLM__REALTIME_CMD_ACCEPT_COUNT' / BitsInteger(8),
    'SOH_COMMAND_TLM__REALTIME_CMD_REJECT_COUNT' / BitsInteger(8),
    'SOH_COMMAND_TLM__STORED_CMD_ACCEPT_COUNT' / BitsInteger(8),
    'SOH_COMMAND_TLM__STORED_CMD_REJECT_COUNT' / BitsInteger(8),
    'SOH_COMMAND_TLM__MACROS_EXECUTING_PACK1' / BitsInteger(8),
    'SOH_COMMAND_TLM__MACROS_EXECUTING_PACK2' / BitsInteger(8),
)

SOH_GENERAL = BitStruct(
    'SOH_GENERAL__SCRUB_STATUS_OVERALL' / Enum(BitsInteger(8), OK=0, FAIL=-1),
    'SOH_GENERAL__IMAGE_BOOTED' / Enum(BitsInteger(8), PRIMARY=0, REDUNDANT=1),
    'SOH_GENERAL__IMAGE_AUTO_FAILOVER' / Enum(BitsInteger(8), OK=0, FAIL=1),
)

SOH_TIME = BitStruct(
    'SOH_TIME__TAI_SECONDS' / LinearAdapter(5.0, BitsInteger(32)),
    'SOH_TIME__TIME_VALID' / Enum(BitsInteger(8), YES=1, NO=0),
)

SOH_REFS = BitStruct(
    'SOH_REFS__POSITION_WRT_ECI1' / LinearAdapter(50000, BitsInteger(32)),
    'SOH_REFS__POSITION_WRT_ECI2' / LinearAdapter(50000, BitsInteger(32)),
    'SOH_REFS__POSITION_WRT_ECI3' / LinearAdapter(50000, BitsInteger(32)),
    'SOH_REFS__VELOCITY_WRT_ECI1' / LinearAdapter(200000000, BitsInteger(32)),
    'SOH_REFS__VELOCITY_WRT_ECI2' / LinearAdapter(200000000, BitsInteger(32)),
    'SOH_REFS__VELOCITY_WRT_ECI3' / LinearAdapter(200000000, BitsInteger(32)),
    'SOH_REFS__REFS_VALID' / Enum(BitsInteger(8), YES=1, NO=0),
)

SOH_ATT_DET = BitStruct(
    'SOH_ATT_DET__Q_BODY_WRT_ECI1' / LinearAdapter(2e9, BitsInteger(32)),
    'SOH_ATT_DET__Q_BODY_WRT_ECI2' / LinearAdapter(2e9, BitsInteger(32)),
    'SOH_ATT_DET__Q_BODY_WRT_ECI3' / LinearAdapter(2e9, BitsInteger(32)),
    'SOH_ATT_DET__Q_BODY_WRT_ECI4' / LinearAdapter(2e9, BitsInteger(32)),
    'SOH_ATT_DET__BODY_RATE1' / LinearAdapter(200000000, BitsInteger(32)),
    'SOH_ATT_DET__BODY_RATE2' / LinearAdapter(200000000, BitsInteger(32)),
    'SOH_ATT_DET__BODY_RATE3' / LinearAdapter(200000000, BitsInteger(32)),
    'SOH_ATT_DET__BAD_ATT_TIMER' / BitsInteger(32),
    'SOH_ATT_DET__BAD_RATE_TIMER' / BitsInteger(32),
    'SOH_ATT_DET__REINIT_COUNT' / BitsInteger(32),
    'SOH_ATT_DET__ATTITUDE_VALID' / Enum(BitsInteger(8), YES=1, NO=0),
    'SOH_ATT_DET__MEAS_ATT_VALID' / Enum(BitsInteger(8), YES=1, NO=0),
    'SOH_ATT_DET__MEAS_RATE_VALID' / Enum(BitsInteger(8), YES=1, NO=0),
    'SOH_ATT_DET__TRACKER_USED' / BitsInteger(8),
)

SOH_ATT_CMD = BitStruct(
    'SOH_ATT_CMD__HR_CYCLE_SAFE_MODE' / BitsInteger(32),
    'SOH_ATT_CMD__ROTISSERIE_RATE' / LinearAdapter(25000, BitsInteger(16)),
    'SOH_ATT_CMD__ADCS_MODE' / Enum(BitsInteger(8), SUN_POINT=0,
                                    FINE_REF_POINT=1, SEARCH_INIT=2,
                                    SEARCHING=3, WAITING=4, CONVERGING=5,
                                    ON_SUN=6, NOT_ACTIVE=7),
    'SOH_ATT_CMD__SAFE_MODE_REASON' / Enum(BitsInteger(8), BOOT=1, COMMAND=2,
                                           ATTITUDE_INVALID=4, TIME_INVALID=8,
                                           ATTITUDE_TIME_INVALID=12,
                                           REFS_INVALID=16),
    'SOH_ATT_CMD__RECOMMEND_SUN_POINT' / Enum(BitsInteger(8), YES=1, NO=0),
)

SOH_RW_DRIVE = BitStruct(
    'SOH_RW_DRIVE__FILTERED_SPEED_RPM1' / LinearAdapter(2.5, BitsInteger(16)),
    'SOH_RW_DRIVE__FILTERED_SPEED_RPM2' / LinearAdapter(2.5, BitsInteger(16)),
    'SOH_RW_DRIVE__FILTERED_SPEED_RPM3' / LinearAdapter(2.5, BitsInteger(16)),
)

SOH_TRACKER = BitStruct(
    'SOH_TRACKER__OPERATING_MODE' / Enum(BitsInteger(8), IDLE=0, INITIALIZE=1,
                                         STARID=2, TRACK=3, PHOTO=4, CAL=5),
    'SOH_TRACKER__STAR_ID_STEP' / Enum(BitsInteger(8), IDLE=0, INITIALIZE=1,
                                       WAITING_FOR_IMAGE1=2,
                                       WAITING_FOR_IMAGE2=3, CALCULATE_RATE=4,
                                       MAKE_UNIT_VECTORS=5, AWAITING_TRISTAR=6,
                                       OK_FOUND_4=7, OK_FOUND_3=8, TIME_OUT=9,
                                       SPARE=10, NO_MATCH=11),
    'SOH_TRACKER__ATT_STATUS' / Enum(BitsInteger(8), OK=0, PENDING=1, BAD=2,
                                     TOO_FEW_STARS=3, QUEST_FAILED=4,
                                     RESIDUALS_TOO_HIGH=5, TOO_CLOSE_TO_EDGE=6,
                                     PIX_AMP_TOO_LOW=7, PIX_AMP_TOO_HIGH=8,
                                     BACKGND_TOO_HIGH=9, TRACK_FAILURE=10,
                                     PIX_SUM_TOO_LOW=11, UNUSED=12,
                                     TOO_DIM_FOR_STARID=13, TOO_MANY_GROUPS=14,
                                     TOO_FEW_GROUPS=15, CHANNEL_DISABLED=16,
                                     TRACK_BLK_OVERLAP=17, OK_FOR_STARID=18,
                                     TOO_CLOSE_TO_OTHER=19, TOO_MANY_PIXELS=20,
                                     TOO_MANY_COLUMNS=21, TOO_MANY_ROWS=22,
                                     OPEN=23, CLOSED=24, RATE_TOO_HIGH=25),
    'SOH_TRACKER__NUM_ATTITUDE_STARS' / BitsInteger(8),
)

SOH_ATT_CTRL = BitStruct(
    'SOH_ATT_CTRL__POSITION_ERROR1' / LinearAdapter(5e8, BitsInteger(32)),
    'SOH_ATT_CTRL__POSITION_ERROR2' / LinearAdapter(5e8, BitsInteger(32)),
    'SOH_ATT_CTRL__POSITION_ERROR3' / LinearAdapter(5e8, BitsInteger(32)),
    'SOH_ATT_CTRL__TIME_INTO_SEARCH' / BitsInteger(16),
    'SOH_ATT_CTRL__WAIT_TIMER' / BitsInteger(16),
    'SOH_ATT_CTRL__SUN_POINT_ANGLE_ERROR' / LinearAdapter(333.3334,
                                                          BitsInteger(16)),
    'SOH_ATT_CTRL__SUN_POINT_STATE' / Enum(BitsInteger(8), SUN_POINT=0,
                                           FINE_REF_POINT=1, SEARCH_INIT=2,
                                           SEARCHING=3, WAITING=4,
                                           CONVERGING=5, ON_SUN=6,
                                           NOT_ACTIVE=7),
)

SOH_MOMENTUM = BitStruct(
    'SOH_MOMENTUM__MOMENTUM_VECTOR_BODY1' / LinearAdapter(5000,
                                                          BitsInteger(16)),
    'SOH_MOMENTUM__MOMENTUM_VECTOR_BODY2' / LinearAdapter(5000,
                                                          BitsInteger(16)),
    'SOH_MOMENTUM__MOMENTUM_VECTOR_BODY3' / LinearAdapter(5000,
                                                          BitsInteger(16)),
    'SOH_MOMENTUM__DUTY_CYCLE1' / BitsInteger(8),
    'SOH_MOMENTUM__DUTY_CYCLE2' / BitsInteger(8),
    'SOH_MOMENTUM__DUTY_CYCLE3' / BitsInteger(8),
    'SOH_MOMENTUM__TORQUE_ROD_MODE1' / Enum(BitsInteger(8), OFF=0, ON_POS=1,
                                            ON_NEG=2, AUTO=3, MEASURED=4,
                                            MODELED=5, DELAYED_AUTO=6,
                                            NO_FIELD_VALID=7),
    'SOH_MOMENTUM__TORQUE_ROD_MODE2' / Enum(BitsInteger(8), OFF=0, ON_POS=1,
                                            ON_NEG=2, AUTO=3, MEASURED=4,
                                            MODELED=5, DELAYED_AUTO=6,
                                            NO_FIELD_VALID=7),
    'SOH_MOMENTUM__TORQUE_ROD_MODE3' / Enum(BitsInteger(8), OFF=0, ON_POS=1,
                                            ON_NEG=2, AUTO=3, MEASURED=4,
                                            MODELED=5, DELAYED_AUTO=6,
                                            NO_FIELD_VALID=7),
    'SOH_MOMENTUM__MAG_SOURCE_USED' / Enum(BitsInteger(8), OFF=0, ON_POS=1,
                                           ON_NEG=2, AUTO=3, MEASURED=4,
                                           MODELED=5, DELAYED_AUTO=6,
                                           NO_FIELD_VALID=7),
    'SOH_MOMENTUM__MOMENTUM_VECTOR_VALID' / Enum(BitsInteger(8), YES=1, NO=0),
)

SOH_CSS = BitStruct(
    'SOH_CSS__SUN_VECTOR_BODY1' / LinearAdapter(10000, BitsInteger(16)),
    'SOH_CSS__SUN_VECTOR_BODY2' / LinearAdapter(10000, BitsInteger(16)),
    'SOH_CSS__SUN_VECTOR_BODY3' / LinearAdapter(10000, BitsInteger(16)),
    'SOH_CSS__SUN_VECTOR_STATUS' / Enum(BitsInteger(8), GOOD=0, COARSE=1,
                                        BAD=2),
    'SOH_CSS__SUN_SENSOR_USED' / BitsInteger(8),
)

SOH_MAG = BitStruct(
    'SOH_MAG__MAG_VECTOR_BODY1' / LinearAdapter(200000000, BitsInteger(16)),
    'SOH_MAG__MAG_VECTOR_BODY2' / LinearAdapter(200000000, BitsInteger(16)),
    'SOH_MAG__MAG_VECTOR_BODY3' / LinearAdapter(200000000, BitsInteger(16)),
    'SOH_MAG__MAG_VECTOR_VALID' / Enum(BitsInteger(8), YES=1, NO=0),
)

SOH_IMU = BitStruct(
    'SOH_IMU__NEW_PACKET_COUNT' / BitsInteger(8),
    'SOH_IMU__IMU_VECTOR_VALID' / Enum(BitsInteger(8), YES=1, NO=0),
)

SOH_CLOCK_SYNC = BitStruct(
    'SOH_CLOCK_SYNC__HR_RUN_COUNT' / BitsInteger(32),
    'SOH_CLOCK_SYNC__HR_EXEC_TIME_MS' / BitsInteger(8),
)

SOH_ANALOGS = BitStruct(
    'SOH_ANALOGS__BOX1_TEMP' / LinearAdapter(200, BitsInteger(16)),
    'SOH_ANALOGS__BUS_VOLTAGE' / LinearAdapter(1000.0, BitsInteger(16)),
    'SOH_ANALOGS__BATTERY_VOLTAGE' / LinearAdapter(500, BitsInteger(16)),
    'SOH_ANALOGS__BATTERY_CURRENT' / LinearAdapter(500, BitsInteger(16)),
)

SOH_TRACKER2 = BitStruct(
    'SOH_TRACKER2__OPERATING_MODE' / Enum(BitsInteger(8), IDLE=0, INITIALIZE=1,
                                          STARID=2, TRACK=3, PHOTO=4, CAL=5),
    'SOH_TRACKER2__STAR_ID_STEP' / Enum(BitsInteger(8), IDLE=0, INITIALIZE=1,
                                        WAITING_FOR_IMAGE1=2,
                                        WAITING_FOR_IMAGE2=3,
                                        CALCULATE_RATE=4,
                                        MAKE_UNIT_VECTORS=5,
                                        AWAITING_TRISTAR=6, OK_FOUND_4=7,
                                        OK_FOUND_3=8, TIME_OUT=9, SPARE=10,
                                        NO_MATCH=11),
    'SOH_TRACKER2__ATT_STATUS' / Enum(BitsInteger(8), OK=0, PENDING=1, BAD=2,
                                      TOO_FEW_STARS=3, QUEST_FAILED=4,
                                      RESIDUALS_TOO_HIGH=5,
                                      TOO_CLOSE_TO_EDGE=6,
                                      PIX_AMP_TOO_LOW=7, PIX_AMP_TOO_HIGH=8,
                                      BACKGND_TOO_HIGH=9, TRACK_FAILURE=10,
                                      PIX_SUM_TOO_LOW=11, UNUSED=12,
                                      TOO_DIM_FOR_STARID=13,
                                      TOO_MANY_GROUPS=14, TOO_FEW_GROUPS=15,
                                      CHANNEL_DISABLED=16,
                                      TRACK_BLK_OVERLAP=17, OK_FOR_STARID=18,
                                      TOO_CLOSE_TO_OTHER=19,
                                      TOO_MANY_PIXELS=20, TOO_MANY_COLUMNS=21,
                                      TOO_MANY_ROWS=22, OPEN=23, CLOSED=24,
                                      RATE_TOO_HIGH=25),
    'SOH_TRACKER2__NUM_ATTITUDE_STARS' / BitsInteger(8),
)

SOH_GPS = BitStruct(
    'SOH_GPS__GPS_LOCK_COUNT' / BitsInteger(16),
    'SOH_GPS__GPS_VALID' / Enum(BitsInteger(8), YES=1, NO=0),
    'SOH_GPS__GPS_ENABLE' / Enum(BitsInteger(8), YES=1, NO=0),
)

SOH_EVENT_CHECK = BitStruct(
    'SOH_EVENT_CHECK__LATCHED_RESP_FIRE_PACK1' / BitsInteger(8),
    'SOH_EVENT_CHECK__LATCHED_RESP_FIRE_PACK2' / BitsInteger(8),
)

SOH_RADIO = BitStruct(
    'SOH_RADIO__SD_MINUTE_CUR' / BitsInteger(32),
    'SOH_RADIO__SD_PERCENT_USED_TOTAL' / BitsInteger(8),
    'SOH_RADIO__SD_OK' / Enum(BitsInteger(8), YES=1, NO=0),
    'SOH_RADIO__SD_FAULT_COUNT' / BitsInteger(8),
    'SOH_RADIO__SDR_TX_TX_FRAMES' / BitsInteger(32),
    'SOH_RADIO__SDR_TX_TEMP' / BitsInteger(8),
    'SOH_RADIO__SDR_TX_COMM_ERROR' / Enum(BitsInteger(8), NO=0, YES=1),
    'SOH_RADIO__SQ_CHANNEL' / BitsInteger(8),
    'SOH_RADIO__SQ_TRAP_COUNT' / BitsInteger(8),
    'SOH_RADIO__SQ_TEMP' / BitsInteger(8),
)

SOH_TRACKER_CTRL = BitStruct(
    'SOH_TRACKER_CTRL__AID_STATUS1' / BitsInteger(8),
    'SOH_TRACKER_CTRL__AID_STATUS2' / BitsInteger(8),
    'SOH_TRACKER_CTRL__STAR_ID_STATUS' / BitsInteger(8)
)

cute_bct_soh = Struct(
    'SOH_L0' / SOH_L0,
    'SOH_COMMAND_TLM' / SOH_COMMAND_TLM,
    'SOH_GENERAL' / SOH_GENERAL,
    'SOH_TIME' / SOH_TIME,
    'SOH_REFS' / SOH_REFS,
    'SOH_ATT_DET' / SOH_ATT_DET,
    'SOH_ATT_CMD' / SOH_ATT_CMD,
    'SOH_RW_DRIVE' / SOH_RW_DRIVE,
    'SOH_TRACKER' / SOH_TRACKER,
    'SOH_ATT_CTRL' / SOH_ATT_CTRL,
    'SOH_MOMENTUM' / SOH_MOMENTUM,
    'SOH_CSS' / SOH_CSS,
    'SOH_MAG' / SOH_MAG,
    'SOH_IMU' / SOH_IMU,
    'SOH_CLOCK_SYNC' / SOH_CLOCK_SYNC,
    'SOH_ANALOGS' / SOH_ANALOGS,
    'SOH_TRACKER2' / SOH_TRACKER2,
    'SOH_GPS' / SOH_GPS,
    'SOH_EVENT_CHECK' / SOH_EVENT_CHECK,
    'SOH_RADIO' / SOH_RADIO,
    'SOH_TRACKER_CTRL' / SOH_TRACKER_CTRL,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Athanasios Theocharis <athatheoc@gmail.com>
# This was made under ESA Summer of Code in Space 2019
# by Athanasios Theocharis, mentored by Daniel Estevez
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
# 

import numpy
from gnuradio import gr
import pmt
from . import telecommand

class telecommand_parser(gr.basic_block):
    """
    docstring for block CCSDS telemetry_parser
    """
    def __init__(self):
        gr.basic_block.__init__(self,
            name="telecommand_parser",
            in_sig=[],
            out_sig=[])

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print("[ERROR] Received invalid message type. Expected u8vector")
            return
        packet = bytearray(pmt.u8vector_elements(msg))

        try:
            data = telecommand.FullPacket.parse(packet[:])
        except:
            print("Could not decode telecommand packet")
            return
        print(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017,2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *


Mode = Enum(
    Int8ub,
    Otherwise=0,
    MissionMode=1,
    )


class BatVoltageAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj - 3.0) * 20.0)

    def _decode(self, obj, context, path=None):
        return obj / 20.0 + 3.0


BatVoltage = BatVoltageAdapter(Int8ub)


class BatCurrentAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj + 1.0) * 127.0)

    def _decode(self, obj, context, path=None):
        return obj / 127.0 - 1.0


BatCurrent = BatCurrentAdapter(Int8ub)


class BusCurrentAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int(obj * 40.0)

    def _decode(self, obj, context, path=None):
        return obj / 40.0


BusCurrent = BusCurrentAdapter(Int8ub)


class TempAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj + 15.0) * 4.0)

    def _decode(self, obj, context, path=None):
        return obj / 4.0 - 15.0


Temp = TempAdapter(Int8ub)

SpacecraftMode = Enum(
    Int8ub,
    GndDebug=0,
    Initial=1,
    Commissioning=2,
    Mission=3,
    Comm=4,
    Powersafe=17,
    Detumble=18,
    Debug=19,
    Standby=32,
    )

kr01 = Struct(
    'header' / Bytes(0x23),
    'mode' / Mode,
    'batvoltage' / BatVoltage,
    'batcurrent' / BatCurrent,
    '3v3buscurrent' / BusCurrent,
    '5vbuscurrent' / BusCurrent,
    'tempcomm' / Temp,
    'tempeps' / Temp,
    'tempbattery' / Temp,
    'spacecraftmode' / SpacecraftMode
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020-2021 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
GNU Radio gr-satellites out-of-tree module.

gr-satellites is a GNU Radio out-of-tree module encompassing a
collection of telemetry decoders that supports many different Amateur
satellites. It suports most popular protocols, such as AX.25, the
GOMspace NanoCom U482C and AX100 modems, an important part of the CCSDS
stack, the AO-40 protocol used in the FUNcube satellites, and several
ad-hoc protocols used in other satellites.
"""

__version__ = 'v3.11.0'
__author__ = 'Daniel Estevez'
__copyright__ = 'Copyright 2016-2021 Daniel Estevez'
__email__ = 'daniel@destevez.net'
__license__ = 'GPL-3.0'
__all__ = [
    'ccsds',
    'components',
    'core',
    'filereceiver',
    'hier',
    'satyaml',
    'telemetry',
    'usp',
    'utils',
    ]

# import swig generated symbols into the satellites namespace
# this would fail if we are being imported from the build dir
try:
    from .satellites_swig import *
except ImportError:
    from satellites_swig import *

# Import any pure python here

from .aausat4_remove_fsm import aausat4_remove_fsm
from .adsb_kml import adsb_kml
from .append_crc32c import append_crc32c
from .beesat_classifier import beesat_classifier
from .bme_submitter import bme_submitter
from .cc11xx_packet_crop import cc11xx_packet_crop
from .check_address import check_address
from .check_ao40_uncoded_crc import check_ao40_uncoded_crc
from .check_astrocast_crc import check_astrocast_crc
from .check_cc11xx_crc import check_cc11xx_crc
from .check_crc16_ccitt import check_crc16_ccitt
from .check_crc16_ccitt_false import check_crc16_ccitt_false
from .check_crc import check_crc
from .check_eseo_crc import check_eseo_crc
from .check_swiatowid_crc import check_swiatowid_crc
from .check_tt64_crc import check_tt64_crc
from .eseo_line_decoder import eseo_line_decoder
from .eseo_packet_crop import eseo_packet_crop
from .fixedlen_tagger import fixedlen_tagger
from .funcube_submit import funcube_submit
from .hdlc_deframer import hdlc_deframer
from .hdlc_framer import hdlc_framer
from .k2sat_deframer import k2sat_deframer
from .kiss_to_pdu import kiss_to_pdu
from .ks1q_header_remover import ks1q_header_remover
from .lilacsat1_gps_kml import lilacsat1_gps_kml
from .manchester_sync import manchester_sync
from .ngham_check_crc import ngham_check_crc
from .ngham_packet_crop import ngham_packet_crop
from .ngham_remove_padding import ngham_remove_padding
from .pdu_to_kiss import pdu_to_kiss
from .print_header import print_header
from .print_timestamp import print_timestamp
from .pwsat2_submitter import pwsat2_submitter
from .pwsat2_telemetry_parser import pwsat2_telemetry_parser
from .reflect_bytes import reflect_bytes
from .snet_classifier import snet_classifier
from .snet_deframer import snet_deframer
from .submit import submit
from .swap_crc import swap_crc
from .swap_header import swap_header
from .swiatowid_packet_crop import swiatowid_packet_crop
from .swiatowid_packet_split import swiatowid_packet_split
from .sx12xx_check_crc import sx12xx_check_crc
from .sx12xx_packet_crop import sx12xx_packet_crop



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import argparse
import shlex


class options_block:
    """
    Class used to add default option parsing to blocks

    A block using this class should inherit both from a GNU
    Radio block class (gr.hier_block2, for example) and options_block.
    It should have a @classmethod add_options(cls, parser) that
    adds class specific options to an argparse parser.

    It should also have an options = None parameter in its __init__()
    method. The __init__() method should call options_block.__init__(options).
    Then options should be access as self.options.

    This class automatically set options to default values even if no options
    arguments was passed to the child class __init__() method.

    Args:
        options: options from argparse
    """
    def __init__(self, options):
        if options is not None and type(options) is not str:
            self.options = options
            return

        p = argparse.ArgumentParser(prog=self.__class__.__name__)
        self.add_options(p)
        self.options = p.parse_args(shlex.split(options) if options else [])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *

from ..adapters import LinearAdapter, AffineAdapter
from .ax25 import Header as AX25Header


BeaconA = BitStruct(
    Bytewise(Const(b'\x00')),  # start byte
    'obc_boot_image' / BitsInteger(8),
    'onboard_time' / BitsInteger(32),
    'uptime' / BitsInteger(32),
    'spacecraft_mode' / BitsInteger(3),
    'separation_seq_state' / BitsInteger(4),
    'solar_array_deploy' / BitsInteger(4)[4],
    'antenna_deploy' / BitsInteger(16)[8],
    'adm_soft_fire_count' / BitsInteger(8),
    'adm_hard_fire_count' / BitsInteger(8),
    'adm_tlm' / BitsInteger(80)[10],
    'sadm_check_count' / BitsInteger(5),
    'sadm_telemetry' / BitsInteger(64)[10],
    'i2c_nack_addr_count' / BitsInteger(32),
    'i2c_hw_state_err_count' / BitsInteger(32),
    'i2c_isr_err_count' / BitsInteger(32),
    'battery_current_dir' / Enum(BitsInteger(1), discharging=0, charging=1),
    'battery_current_5v' / LinearAdapter(1/1.327547, BitsInteger(10)),
    'battery_current_msb_3v3' / BitsInteger(1)
    )


class SignMagnitudeAdapter16(Adapter):
    def _encode(self, obj, context, path=None):
        mag = abs(obj) & 0x7fff
        sig = 1 if obj < 0 else 0
        return (sig << 15) | mag

    def _decode(self, obj, context, path=None):
        mag = obj & 0x7fff
        return -mag if obj & 0x8000 else mag


SignMagnitude16 = SignMagnitudeAdapter16(Int16ub)


class SignMagnitudeAdapter8(Adapter):
    def _encode(self, obj, context, path=None):
        mag = abs(obj) & 0x7f
        sig = 1 if obj < 0 else 0
        return (sig << 7) | mag

    def _decode(self, obj, context, path=None):
        mag = obj & 0x7f
        return -mag if obj & 0x80 else mag


SignMagnitude8 = SignMagnitudeAdapter8(Int8ub)

BeaconPartB = BitStruct(
    'battery_current_lsbs_3v3' / BitsInteger(9),
    'battery_current_vbat' / LinearAdapter(1/14.662757, BitsInteger(10)),
    'battery_voltage_3v3' / LinearAdapter(1/0.004311, BitsInteger(10)),
    'battery_voltage_5v' / LinearAdapter(1/0.005865, BitsInteger(10)),
    'battery_voltage_vbat' / LinearAdapter(1/0.008993, BitsInteger(10)),
    'battery_temp' / AffineAdapter(1/0.3976, 238.57/0.3976,
                                   BitsInteger(10)),
    'solar_current' / LinearAdapter(1/0.977517107, BitsInteger(10)),
    # Last 5 solar_voltage's unused
    'solar_voltage' / LinearAdapter(1/0.009971, BitsInteger(10))[9],
    'eps_bus_voltage_vbat' / LinearAdapter(1/0.008978, BitsInteger(10)),
    'eps_bus_voltage_3v3' / LinearAdapter(1/0.00431085, BitsInteger(10)),
    'eps_bus_voltage_5v' / LinearAdapter(1/0.005865103, BitsInteger(10)),
    'eps_bus_voltage_12v' / LinearAdapter(1/0.013489736, BitsInteger(10)),
    'eps_bus_current_vbat' / LinearAdapter(1/2.07, BitsInteger(10)),
    'eps_bus_current_3v3' / LinearAdapter(1/5.236698785, BitsInteger(10)),
    'eps_bus_current_5v' / LinearAdapter(1/5.236698785, BitsInteger(10)),
    'eps_bus_current_12v' / LinearAdapter(1/2.07, BitsInteger(10)),
    'adcs_raw_gyro' / Bytewise(LinearAdapter(-1/0.0104166,
                                             SignMagnitude16)[3]),
    'adcs_mtq_dir_duty' / BitsInteger(8)[6],
    'adcs_status' / BitsInteger(16),
    'adcs_bus_voltage_5v' / LinearAdapter(1/0.000152587, BitsInteger(16)),
    'adcs_bus_voltage_3v3' / LinearAdapter(1/0.0000944, BitsInteger(16)),
    'adcs_bus_voltage_1v5' / LinearAdapter(1/0.000152587, BitsInteger(16)),
    'adcs_bus_current_5v' / LinearAdapter(1/0.00000600472, BitsInteger(16)),
    'adcs_bus_current_3v3' / LinearAdapter(1/0.0000381, BitsInteger(16)),
    'adcs_bus_current_1v5' / LinearAdapter(1/0.0000244230769230769,
                                           BitsInteger(16)),
    'adcs_board_temp' / AffineAdapter(
        1/0.00762, 273.150/0.00762, BitsInteger(16)),
    'adcs_adc_ref' / BitsInteger(16),
    'adcs_sensor_current' / LinearAdapter(1/0.0000015698342656893,
                                          BitsInteger(16)),
    'adcs_mtq_current' / LinearAdapter(1/0.00000601213123029945,
                                       BitsInteger(16)),
    'adcs_array_temp' / AffineAdapter(1/0.00762, 273.150/0.00762,
                                      BitsInteger(16))[6],
    'adcs_css_raw' / LinearAdapter(1/0.0415032679738562,
                                   BitsInteger(16))[6],
    'fss_active' / BitsInteger(2)[6],
    'css_active_selected' / BitsInteger(2)[6],
    'adcs_sun_processed' / AffineAdapter(
        1/0.000030517578125, 1/0.000030517578125, BitsInteger(16))[3],
    'reserved' / BitsInteger(16)[4],
    'adcs_detumble_counter' / BitsInteger(16),
    'adcs_mode' / Enum(BitsInteger(16), standby=0, detumble=1,
                       coarse_point=2, fine_point=3),
    'adcs_state' / Enum(BitsInteger(16), nadir=0, sun=1, velocity=2,
                        LLA=3, moon=4),
    'reservedA' / BitsInteger(10),
    'reservedB' / BitsInteger(4),
    'reservedC' / BitsInteger(16),
    'reservedD' / BitsInteger(3),
    'reservedE' / BitsInteger(4),
    'cmc_rx_lock' / Flag,
    'cmc_rx_frame_count' / BitsInteger(16),
    'cmc_rx_packet_count' / BitsInteger(16),
    'cmc_rx_dropped_error_count' / BitsInteger(16),
    'cmc_rx_crc_error_count' / BitsInteger(16),
    'cmc_rx_overrun_error_count' / BitsInteger(8),
    'cmc_rx_protocol_error_count' / BitsInteger(16),
    'cmc_smps_temperature' / Bytewise(SignMagnitude8),
    'cmc_pa_temperature' / Bytewise(SignMagnitude8),
    'ax25_mux_channel_enable' / Flag[3],
    'digipeater_enable' / Flag,
    'pacsat_broadcast_enable' / Flag,
    'pacsat_broadcast_in_progress' / Flag,
    'param_valid_flags' / Flag[40],
    Padding(5),
    'checksum' / BitsInteger(16),
)

BeaconHeader = Struct(
    'packet_type' / Int8ub,
    'apid' / Int8ub,
    'sequence_count' / Int16ub,
    'length' / Int16ub,
    'reserved' / Int8ub,
    'service_type' / Const(b'\x03'),
    'service_subtype' / Const(b'\x19')
    )

BeaconPartA = Struct(
    'beacon_header' / BeaconHeader,
    'beacon' / BeaconA,
    )

mirsat1 = Struct(
    'ax25_header' / AX25Header,
    'telemetry' / If(
        this.ax25_header.pid == 0xF0,
        Struct(
            'header' / Bytes(6),
            'beacon' / Select(BeaconPartA, BeaconPartB)
            ))
    )



#!D:\bld\gnuradio-satellites_1642173413548\_h_env\python.exe
# -*- coding: utf-8 -*-

# Copyright 2019-2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import subprocess
import sys

import numpy as np

from satellites.kiss import *


def print_usage():
    print(f'Usage {sys.argv[0]} <jy1sat_frames.kss> <output_path>')


def seqnum(packet):
    return packet[3]*256 + packet[4]


def read_kiss_file(path):
    frames = list()
    frame = list()
    framesize = 256
    transpose = False
    with open(path, 'rb') as f:
        for c in f.read():
            if c == FEND:
                if len(frame) == framesize + 1 and (frame[0] & 0x0f) == 0:
                    frames.append(frame[1:])
                frame = list()
            elif transpose:
                if c == TFEND:
                    frame.append(FEND)
                elif c == TFESC:
                    frame.append(FESC)
                transpose = False
            elif c == FESC:
                transpose = True
            else:
                frame.append(c)
    return np.array(frames, dtype='uint8')


def main():
    if len(sys.argv) != 3:
        print_usage()
        sys.exit(1)

    input_file = sys.argv[1]
    output_file = sys.argv[2]

    # Read 256 byte frames
    x = read_kiss_file(input_file)

    # Filter out by frame id and trim to payload
    x = x[((x[:, 0] == 0xe0) | (x[:, 0] == 0xe1)) & (x[:, 1] == 0x10), 56:]

    # Filter SSDV packets
    x = x[(x[:, 0] == 0x55) & (x[:, 1] == 0x68), :]
    ids = set(x[:, 2])

    for i in ids:
        L = list(x[x[:, 2] == i, :])
        L.sort(key=seqnum)
        ssdv = '{}_{}.ssdv'.format(output_file, i)
        jpeg = '{}_{}.jpg'.format(output_file, i)
        np.array(L).tofile(ssdv)
        print('Calling SSDV decoder for image {}'.format(hex(i)))
        subprocess.call(['ssdv', '-d', '-J', ssdv, jpeg])
        print()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr
import numpy
import pmt

from . import csp_header


class print_header(gr.basic_block):
    """Print the CSP header of a CSP packet"""
    def __init__(self):
        gr.basic_block.__init__(
            self,
            name='print_header',
            in_sig=[],
            out_sig=[])

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = bytes(pmt.u8vector_elements(msg))
        try:
            print(csp_header.CSP(packet[:4]))
        except ValueError as e:
            print(e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import importlib

from gnuradio import gr, blocks, digital


class mobitex_deframer(gr.hier_block2):
    """
    Hierarchical block to deframe Mobitex and Mobitex-NX

    The input is a float stream of soft symbols. The output are PDUs
    with Mobitex-NX frames.

    This uses the gr-tnc_nx OOT module, but only attempts to import
    it when __init__() is called.

    Args:
        nx: use NX mode (boolean)
        options: Options from argparse
    """
    def __init__(self, nx=False, options=None):
        gr.hier_block2.__init__(
            self,
            'mobitex_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        self.message_port_register_hier_out('out')

        self.invert = blocks.multiply_const_ff(-1, 1)
        self.slicer = digital.binary_slicer_fb()
        try:
            tnc_nx = importlib.import_module('tnc_nx')
        except ImportError as e:
            print('Unable to import tnc_nx')
            print('gr-tnc_nx needs to be installed to use Mobitex')
            raise e
        syncword = 0x0ef0 if nx else 0x5765
        self.mobitex = tnc_nx.nx_decoder(syncword, nx)

        self.connect(self, self.invert, self.slicer, self.mobitex)
        self.msg_connect((self.mobitex, 'out'), (self, 'out'))
